# Copyright 2012 NetApp
# Copyright 2014 Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""Unit tests for the Share driver module."""

import time
from unittest import mock


import ddt

from manila.common import constants
from manila import exception
from manila import network
from manila.share import configuration
from manila.share import driver
from manila import test
from manila.tests import utils as test_utils
from manila import utils


def fake_execute_with_raise(*cmd, **kwargs):
    raise exception.ProcessExecutionError


def fake_sleep(duration):
    pass


class ShareDriverWithExecuteMixin(driver.ShareDriver, driver.ExecuteMixin):
    pass


@ddt.ddt
class ShareDriverTestCase(test.TestCase):
    _SNAPSHOT_METHOD_NAMES = ["create_snapshot", "delete_snapshot"]

    def setUp(self):
        super(ShareDriverTestCase, self).setUp()
        self.utils = utils
        self.mock_object(self.utils, 'execute', fake_execute_with_raise)
        self.time = time
        self.mock_object(self.time, 'sleep', fake_sleep)
        driver.CONF.set_default('driver_handles_share_servers', True)

    def test__try_execute(self):
        execute_mixin = ShareDriverWithExecuteMixin(
            True, configuration=configuration.Configuration(None))
        self.assertRaises(exception.ProcessExecutionError,
                          execute_mixin._try_execute)

    def test_verify_share_driver_mode_option_type(self):
        data = {'DEFAULT': {'driver_handles_share_servers': 'True'}}
        with test_utils.create_temp_config_with_opts(data):
            share_driver = driver.ShareDriver([True, False])
            self.assertTrue(share_driver.driver_handles_share_servers)

    def _instantiate_share_driver(self, network_config_group,
                                  driver_handles_share_servers,
                                  admin_network_config_group=None):
        self.mock_object(network, 'API')
        config = mock.Mock()
        config.append_config_values = mock.Mock()
        config.config_group = 'fake_config_group'
        config.network_config_group = network_config_group
        if admin_network_config_group:
            config.admin_network_config_group = admin_network_config_group
        config.safe_get = mock.Mock(return_value=driver_handles_share_servers)

        share_driver = driver.ShareDriver([True, False], configuration=config)

        self.assertTrue(hasattr(share_driver, 'configuration'))
        config.append_config_values.assert_called_once_with(driver.share_opts)
        if driver_handles_share_servers:
            calls = []
            if network_config_group:
                calls.append(mock.call(
                    config_group_name=config.network_config_group))
            else:
                calls.append(mock.call(
                    config_group_name=config.config_group))
            if admin_network_config_group:
                calls.append(mock.call(
                    config_group_name=config.admin_network_config_group,
                    label='admin'))
            network.API.assert_has_calls(calls)
            self.assertTrue(hasattr(share_driver, 'network_api'))
            self.assertTrue(hasattr(share_driver, 'admin_network_api'))
            self.assertIsNotNone(share_driver.network_api)
            self.assertIsNotNone(share_driver.admin_network_api)
        else:
            self.assertFalse(hasattr(share_driver, 'network_api'))
            self.assertTrue(hasattr(share_driver, 'admin_network_api'))
            self.assertIsNone(share_driver.admin_network_api)
            self.assertFalse(network.API.called)
        return share_driver

    def test_instantiate_share_driver(self):
        self._instantiate_share_driver(None, True)

    def test_instantiate_share_driver_another_config_group(self):
        self._instantiate_share_driver("fake_network_config_group", True)

    def test_instantiate_share_driver_with_admin_network(self):
        self._instantiate_share_driver(
            "fake_network_config_group", True,
            "fake_admin_network_config_group")

    def test_instantiate_share_driver_no_configuration(self):
        self.mock_object(network, 'API')

        share_driver = driver.ShareDriver(True, configuration=None)

        self.assertIsNone(share_driver.configuration)
        network.API.assert_called_once_with(config_group_name=None)

    def test_get_share_stats_refresh_false(self):
        share_driver = driver.ShareDriver(True, configuration=None)
        share_driver._stats = {'fake_key': 'fake_value'}

        result = share_driver.get_share_stats(False)

        self.assertEqual(share_driver._stats, result)

    def test_get_share_stats_refresh_true(self):
        conf = configuration.Configuration(None)
        expected_keys = [
            'qos', 'driver_version', 'share_backend_name',
            'free_capacity_gb', 'total_capacity_gb',
            'driver_handles_share_servers',
            'reserved_percentage', 'reserved_snapshot_percentage',
            'vendor_name', 'storage_protocol',
            'snapshot_support', 'mount_snapshot_support',
        ]
        share_driver = driver.ShareDriver(True, configuration=conf)
        fake_stats = {'fake_key': 'fake_value'}
        share_driver._stats = fake_stats

        result = share_driver.get_share_stats(True)

        self.assertNotEqual(fake_stats, result)
        for key in expected_keys:
            self.assertIn(key, result)
        self.assertEqual('Open Source', result['vendor_name'])

    def test_get_share_status(self):
        share_driver = self._instantiate_share_driver(None, False)
        self.assertRaises(NotImplementedError,
                          share_driver.get_share_status,
                          None, None)

    @ddt.data(
        {'opt': True, 'allowed': True},
        {'opt': True, 'allowed': (True, False)},
        {'opt': True, 'allowed': [True, False]},
        {'opt': True, 'allowed': set([True, False])},
        {'opt': False, 'allowed': False},
        {'opt': False, 'allowed': (True, False)},
        {'opt': False, 'allowed': [True, False]},
        {'opt': False, 'allowed': set([True, False])})
    @ddt.unpack
    def test__verify_share_server_handling_valid_cases(self, opt, allowed):
        conf = configuration.Configuration(None)
        self.mock_object(conf, 'safe_get', mock.Mock(return_value=opt))
        share_driver = driver.ShareDriver(allowed, configuration=conf)
        self.assertTrue(conf.safe_get.called)
        self.assertEqual(opt, share_driver.driver_handles_share_servers)

    @ddt.data(
        {'opt': False, 'allowed': True},
        {'opt': True, 'allowed': False},
        {'opt': None, 'allowed': True},
        {'opt': 'True', 'allowed': True},
        {'opt': 'False', 'allowed': False},
        {'opt': [], 'allowed': True},
        {'opt': True, 'allowed': []},
        {'opt': True, 'allowed': ['True']},
        {'opt': False, 'allowed': ['False']})
    @ddt.unpack
    def test__verify_share_server_handling_invalid_cases(self, opt, allowed):
        conf = configuration.Configuration(None)
        self.mock_object(conf, 'safe_get', mock.Mock(return_value=opt))
        self.assertRaises(
            exception.ManilaException,
            driver.ShareDriver, allowed, configuration=conf)
        self.assertTrue(conf.safe_get.called)

    def test_setup_server_handling_disabled(self):
        share_driver = self._instantiate_share_driver(None, False)
        # We expect successful execution, nothing to assert
        share_driver.setup_server('Nothing is expected to happen.')

    def test_setup_server_handling_enabled(self):
        share_driver = self._instantiate_share_driver(None, True)
        self.assertRaises(
            NotImplementedError,
            share_driver.setup_server,
            'fake_network_info')

    def test_teardown_server_handling_disabled(self):
        share_driver = self._instantiate_share_driver(None, False)
        # We expect successful execution, nothing to assert
        share_driver.teardown_server('Nothing is expected to happen.')

    def test_teardown_server_handling_enabled(self):
        share_driver = self._instantiate_share_driver(None, True)
        self.assertRaises(
            NotImplementedError,
            share_driver.teardown_server,
            'fake_share_server_details')

    def _assert_is_callable(self, obj, attr):
        self.assertTrue(callable(getattr(obj, attr)))

    @ddt.data('manage_existing',
              'unmanage')
    def test_drivers_methods_needed_by_manage_functionality(self, method):
        share_driver = self._instantiate_share_driver(None, False)

        self._assert_is_callable(share_driver, method)

    @ddt.data('manage_existing_snapshot',
              'unmanage_snapshot')
    def test_drivers_methods_needed_by_manage_snapshot_functionality(
            self, method):
        share_driver = self._instantiate_share_driver(None, False)

        self._assert_is_callable(share_driver, method)

    @ddt.data('revert_to_snapshot',
              'revert_to_replicated_snapshot')
    def test_drivers_methods_needed_by_share_revert_to_snapshot_functionality(
            self, method):
        share_driver = self._instantiate_share_driver(None, False)

        self._assert_is_callable(share_driver, method)

    @ddt.data(True, False)
    def test_get_share_server_pools(self, value):
        driver.CONF.set_default('driver_handles_share_servers', value)
        share_driver = driver.ShareDriver(value)
        self.assertEqual([],
                         share_driver.get_share_server_pools('fake_server'))

    @ddt.data(0.8, 1.0, 10.5, 20.0, None, '1', '1.1')
    def test_check_for_setup_error(self, value):
        driver.CONF.set_default('driver_handles_share_servers', False)
        share_driver = driver.ShareDriver(False)
        share_driver.configuration = configuration.Configuration(None)
        self.mock_object(share_driver.configuration, 'safe_get',
                         mock.Mock(return_value=value))
        if value and float(value) >= 1.0:
            share_driver.check_for_setup_error()
        else:
            self.assertRaises(exception.InvalidParameterValue,
                              share_driver.check_for_setup_error)

    def test_snapshot_support_exists(self):
        driver.CONF.set_default('driver_handles_share_servers', True)
        fake_method = lambda *args, **kwargs: None  # noqa: E731
        child_methods = {
            "create_snapshot": fake_method,
            "delete_snapshot": fake_method,
        }
        child_class_instance = type(
            "NotRedefined", (driver.ShareDriver, ), child_methods)(True)
        self.mock_object(child_class_instance, "configuration")

        child_class_instance._update_share_stats()

        self.assertTrue(child_class_instance._stats["snapshot_support"])
        self.assertTrue(child_class_instance.configuration.safe_get.called)

    @ddt.data(
        ([], [], False),
        (_SNAPSHOT_METHOD_NAMES, [], True),
        (_SNAPSHOT_METHOD_NAMES, _SNAPSHOT_METHOD_NAMES, True),
        (_SNAPSHOT_METHOD_NAMES[0:1], _SNAPSHOT_METHOD_NAMES[1:], True),
        ([], _SNAPSHOT_METHOD_NAMES, True),
    )
    @ddt.unpack
    def test_check_redefined_driver_methods(self, common_drv_meth_names,
                                            child_drv_meth_names,
                                            expected_result):
        # This test covers the case of drivers inheriting other drivers or
        # common classes.

        driver.CONF.set_default('driver_handles_share_servers', True)

        common_drv_methods, child_drv_methods = [
            {method_name: lambda *args, **kwargs: None  # noqa: E731
             for method_name in method_names}
            for method_names in (common_drv_meth_names,
                                 child_drv_meth_names)]

        common_drv = type(
            "NotRedefinedCommon", (driver.ShareDriver, ), common_drv_methods)
        child_drv_instance = type("NotRedefined", (common_drv, ),
                                  child_drv_methods)(True)

        has_redefined_methods = (
            child_drv_instance._has_redefined_driver_methods(
                self._SNAPSHOT_METHOD_NAMES))

        self.assertEqual(expected_result, has_redefined_methods)

    @ddt.data(
        (),
        ("create_snapshot"),
        ("delete_snapshot"),
        ("create_snapshot", "delete_snapshotFOO"),
    )
    def test_snapshot_support_absent(self, methods):
        driver.CONF.set_default('driver_handles_share_servers', True)
        fake_method = lambda *args, **kwargs: None  # noqa: E731
        child_methods = {}
        for method in methods:
            child_methods[method] = fake_method
        child_class_instance = type(
            "NotRedefined", (driver.ShareDriver, ), child_methods)(True)
        self.mock_object(child_class_instance, "configuration")

        child_class_instance._update_share_stats()

        self.assertFalse(child_class_instance._stats["snapshot_support"])
        self.assertTrue(child_class_instance.configuration.safe_get.called)

    @ddt.data(True, False)
    def test_snapshot_support_not_exists_and_set_explicitly(
            self, snapshots_are_supported):
        driver.CONF.set_default('driver_handles_share_servers', True)
        child_class_instance = type(
            "NotRedefined", (driver.ShareDriver, ), {})(True)
        self.mock_object(child_class_instance, "configuration")

        child_class_instance._update_share_stats(
            {"snapshot_support": snapshots_are_supported})

        self.assertEqual(
            snapshots_are_supported,
            child_class_instance._stats["snapshot_support"])
        self.assertTrue(child_class_instance.configuration.safe_get.called)

    @ddt.data(True, False)
    def test_snapshot_support_exists_and_set_explicitly(
            self, snapshots_are_supported):
        driver.CONF.set_default('driver_handles_share_servers', True)
        fake_method = lambda *args, **kwargs: None  # noqa: E731
        child_methods = {
            "create_snapshot": fake_method,
            "delete_snapshot": fake_method,
        }
        child_class_instance = type(
            "NotRedefined", (driver.ShareDriver, ), child_methods)(True)
        self.mock_object(child_class_instance, "configuration")

        child_class_instance._update_share_stats(
            {"snapshot_support": snapshots_are_supported})

        self.assertEqual(
            snapshots_are_supported,
            child_class_instance._stats["snapshot_support"])
        self.assertTrue(child_class_instance.configuration.safe_get.called)

    def test_create_share_from_snapshot_support_exists(self):
        driver.CONF.set_default('driver_handles_share_servers', True)
        fake_method = lambda *args, **kwargs: None  # noqa: E731
        child_methods = {
            "create_share_from_snapshot": fake_method,
            "create_snapshot": fake_method,
            "delete_snapshot": fake_method,
        }
        child_class_instance = type(
            "NotRedefined", (driver.ShareDriver, ), child_methods)(True)
        self.mock_object(child_class_instance, "configuration")

        child_class_instance._update_share_stats()

        self.assertTrue(
            child_class_instance._stats["create_share_from_snapshot_support"])
        self.assertTrue(child_class_instance.configuration.safe_get.called)

    @ddt.data(
        (),
        ("create_snapshot"),
        ("create_share_from_snapshotFOO"),
    )
    def test_create_share_from_snapshot_support_absent(self, methods):
        driver.CONF.set_default('driver_handles_share_servers', True)
        fake_method = lambda *args, **kwargs: None  # noqa: E731
        child_methods = {}
        for method in methods:
            child_methods[method] = fake_method
        child_class_instance = type(
            "NotRedefined", (driver.ShareDriver, ), child_methods)(True)
        self.mock_object(child_class_instance, "configuration")

        child_class_instance._update_share_stats()

        self.assertFalse(
            child_class_instance._stats["create_share_from_snapshot_support"])
        self.assertTrue(child_class_instance.configuration.safe_get.called)

    @ddt.data(True, False)
    def test_create_share_from_snapshot_not_exists_and_set_explicitly(
            self, creating_shares_from_snapshot_is_supported):
        driver.CONF.set_default('driver_handles_share_servers', True)
        child_class_instance = type(
            "NotRedefined", (driver.ShareDriver, ), {})(True)
        self.mock_object(child_class_instance, "configuration")

        child_class_instance._update_share_stats({
            "create_share_from_snapshot_support":
                creating_shares_from_snapshot_is_supported,
        })

        self.assertEqual(
            creating_shares_from_snapshot_is_supported,
            child_class_instance._stats["create_share_from_snapshot_support"])
        self.assertTrue(child_class_instance.configuration.safe_get.called)

    @ddt.data(True, False)
    def test_create_share_from_snapshot_exists_and_set_explicitly(
            self, create_share_from_snapshot_supported):
        driver.CONF.set_default('driver_handles_share_servers', True)
        fake_method = lambda *args, **kwargs: None  # noqa: E731
        child_methods = {"create_share_from_snapshot": fake_method}
        child_class_instance = type(
            "NotRedefined", (driver.ShareDriver, ), child_methods)(True)
        self.mock_object(child_class_instance, "configuration")

        child_class_instance._update_share_stats({
            "create_share_from_snapshot_support":
                create_share_from_snapshot_supported,
        })

        self.assertEqual(
            create_share_from_snapshot_supported,
            child_class_instance._stats["create_share_from_snapshot_support"])
        self.assertTrue(child_class_instance.configuration.safe_get.called)

    def test_get_periodic_hook_data(self):
        share_driver = self._instantiate_share_driver(None, False)
        share_instances = ["list", "of", "share", "instances"]

        result = share_driver.get_periodic_hook_data(
            "fake_context", share_instances)

        self.assertEqual(share_instances, result)

    def test_get_admin_network_allocations_number(self):
        share_driver = self._instantiate_share_driver(None, True)

        self.assertEqual(
            0, share_driver.get_admin_network_allocations_number())

    def test_allocate_admin_network_count_None(self):
        share_driver = self._instantiate_share_driver(None, True)
        ctxt = 'fake_context'
        share_server = 'fake_share_server'
        mock_get_admin_network_allocations_number = self.mock_object(
            share_driver,
            'get_admin_network_allocations_number',
            mock.Mock(return_value=0))
        self.mock_object(
            share_driver.admin_network_api,
            'allocate_network',
            mock.Mock(side_effect=Exception('ShouldNotBeRaised')))

        share_driver.allocate_admin_network(ctxt, share_server)

        mock_get_admin_network_allocations_number.assert_called_once_with()
        self.assertFalse(
            share_driver.admin_network_api.allocate_network.called)

    def test_allocate_admin_network_count_0(self):
        share_driver = self._instantiate_share_driver(None, True)
        ctxt = 'fake_context'
        share_server = 'fake_share_server'
        self.mock_object(
            share_driver,
            'get_admin_network_allocations_number',
            mock.Mock(return_value=0))
        self.mock_object(
            share_driver.admin_network_api,
            'allocate_network',
            mock.Mock(side_effect=Exception('ShouldNotBeRaised')))

        share_driver.allocate_admin_network(ctxt, share_server, count=0)

        self.assertFalse(
            share_driver.get_admin_network_allocations_number.called)
        self.assertFalse(
            share_driver.admin_network_api.allocate_network.called)

    def test_allocate_admin_network_count_1_api_initialized(self):
        share_driver = self._instantiate_share_driver(None, True)
        ctxt = 'fake_context'
        share_server = 'fake_share_server'
        mock_get_admin_network_allocations_number = self.mock_object(
            share_driver,
            'get_admin_network_allocations_number',
            mock.Mock(return_value=1))
        self.mock_object(
            share_driver.admin_network_api,
            'allocate_network',
            mock.Mock())

        share_driver.allocate_admin_network(ctxt, share_server)

        mock_get_admin_network_allocations_number.assert_called_once_with()
        (share_driver.admin_network_api.allocate_network.
            assert_called_once_with(ctxt, share_server, count=1))

    def test_allocate_admin_network_count_1_api_not_initialized(self):
        share_driver = self._instantiate_share_driver(None, True, None)
        ctxt = 'fake_context'
        share_server = 'fake_share_server'
        share_driver._admin_network_api = None
        mock_get_admin_network_allocations_number = self.mock_object(
            share_driver,
            'get_admin_network_allocations_number',
            mock.Mock(return_value=1))

        self.assertRaises(
            exception.NetworkBadConfigurationException,
            share_driver.allocate_admin_network,
            ctxt, share_server,
        )
        mock_get_admin_network_allocations_number.assert_called_once_with()

    def test_migration_start(self):

        driver.CONF.set_default('driver_handles_share_servers', False)
        share_driver = driver.ShareDriver(False)

        self.assertRaises(NotImplementedError, share_driver.migration_start,
                          None, None, None, None, None, None, None)

    def test_migration_continue(self):

        driver.CONF.set_default('driver_handles_share_servers', False)
        share_driver = driver.ShareDriver(False)

        self.assertRaises(NotImplementedError, share_driver.migration_continue,
                          None, None, None, None, None, None, None)

    def test_migration_complete(self):

        driver.CONF.set_default('driver_handles_share_servers', False)
        share_driver = driver.ShareDriver(False)

        self.assertRaises(NotImplementedError, share_driver.migration_complete,
                          None, None, None, None, None, None, None)

    def test_migration_cancel(self):

        driver.CONF.set_default('driver_handles_share_servers', False)
        share_driver = driver.ShareDriver(False)

        self.assertRaises(NotImplementedError, share_driver.migration_cancel,
                          None, None, None, None, None, None, None)

    def test_migration_get_progress(self):

        driver.CONF.set_default('driver_handles_share_servers', False)
        share_driver = driver.ShareDriver(False)

        self.assertRaises(NotImplementedError,
                          share_driver.migration_get_progress,
                          None, None, None, None, None, None, None)

    def test_share_server_migration_start(self):
        driver.CONF.set_default('driver_handles_share_servers', True)
        share_driver = driver.ShareDriver(True)

        self.assertRaises(NotImplementedError,
                          share_driver.share_server_migration_start,
                          None, None, None, None, None)

    def test_share_server_migration_continue(self):
        driver.CONF.set_default('driver_handles_share_servers', True)
        share_driver = driver.ShareDriver(True)

        self.assertRaises(NotImplementedError,
                          share_driver.share_server_migration_continue,
                          None, None, None, None, None)

    def test_share_server_migration_get_progress(self):
        driver.CONF.set_default('driver_handles_share_servers', True)
        share_driver = driver.ShareDriver(True)

        self.assertRaises(NotImplementedError,
                          share_driver.share_server_migration_get_progress,
                          None, None, None, None, None)

    def test_share_server_migration_cancel(self):
        driver.CONF.set_default('driver_handles_share_servers', True)
        share_driver = driver.ShareDriver(True)

        self.assertRaises(NotImplementedError,
                          share_driver.share_server_migration_cancel,
                          None, None, None, None, None)

    def test_share_server_migration_check_compatibility(self):
        driver.CONF.set_default('driver_handles_share_servers', True)
        share_driver = driver.ShareDriver(True)
        expected_compatibility = {
            'compatible': False,
            'writable': False,
            'nondisruptive': False,
            'preserve_snapshots': False,
            'migration_cancel': False,
            'migration_get_progress': False,
        }

        driver_compatibility = (
            share_driver.share_server_migration_check_compatibility(
                None, None, None, None, None, None))
        self.assertEqual(expected_compatibility, driver_compatibility)

    def test_share_server_migration_complete(self):
        driver.CONF.set_default('driver_handles_share_servers', True)
        share_driver = driver.ShareDriver(True)

        self.assertRaises(
            NotImplementedError,
            share_driver.share_server_migration_complete,
            None, None, None, None, None, None)

    @ddt.data(True, False)
    def test_connection_get_info(self, admin):

        expected = {
            'mount': 'mount -vt nfs %(options)s /fake/fake_id %(path)s',
            'unmount': 'umount -v %(path)s',
            'access_mapping': {
                'ip': ['nfs']
            }
        }

        fake_share = {
            'id': 'fake_id',
            'share_proto': 'nfs',
            'export_locations': [{
                'path': '/fake/fake_id',
                'is_admin_only': admin
            }]
        }

        driver.CONF.set_default('driver_handles_share_servers', False)
        share_driver = driver.ShareDriver(False)
        share_driver.configuration = configuration.Configuration(None)

        connection_info = share_driver.connection_get_info(
            None, fake_share, "fake_server")

        self.assertEqual(expected, connection_info)

    def test_migration_check_compatibility(self):

        driver.CONF.set_default('driver_handles_share_servers', False)
        share_driver = driver.ShareDriver(False)
        share_driver.configuration = configuration.Configuration(None)
        expected = {
            'compatible': False,
            'writable': False,
            'preserve_metadata': False,
            'nondisruptive': False,
            'preserve_snapshots': False,
        }

        result = share_driver.migration_check_compatibility(
            None, None, None, None, None)

        self.assertEqual(expected, result)

    def test_update_access(self):
        share_driver = driver.ShareDriver(True, configuration=None)
        self.assertRaises(
            NotImplementedError,
            share_driver.update_access,
            'ctx',
            'fake_share',
            'fake_access_rules',
            'fake_add_rules',
            'fake_delete_rules'
        )

    def test_create_replica(self):
        share_driver = self._instantiate_share_driver(None, True)
        self.assertRaises(NotImplementedError,
                          share_driver.create_replica,
                          'fake_context', ['r1', 'r2'],
                          'fake_new_replica', [], [])

    def test_delete_replica(self):
        share_driver = self._instantiate_share_driver(None, True)
        self.assertRaises(NotImplementedError,
                          share_driver.delete_replica,
                          'fake_context', ['r1', 'r2'],
                          'fake_replica', [])

    def test_promote_replica(self):
        share_driver = self._instantiate_share_driver(None, True)
        self.assertRaises(NotImplementedError,
                          share_driver.promote_replica,
                          'fake_context', [], 'fake_replica', [])

    def test_update_replica_state(self):
        share_driver = self._instantiate_share_driver(None, True)
        self.assertRaises(NotImplementedError,
                          share_driver.update_replica_state,
                          'fake_context', ['r1', 'r2'], 'fake_replica', [], [])

    def test_create_replicated_snapshot(self):
        share_driver = self._instantiate_share_driver(None, False)
        self.assertRaises(NotImplementedError,
                          share_driver.create_replicated_snapshot,
                          'fake_context', ['r1', 'r2'], ['s1', 's2'])

    def test_delete_replicated_snapshot(self):
        share_driver = self._instantiate_share_driver(None, False)
        self.assertRaises(NotImplementedError,
                          share_driver.delete_replicated_snapshot,
                          'fake_context', ['r1', 'r2'], ['s1', 's2'])

    def test_update_replicated_snapshot(self):
        share_driver = self._instantiate_share_driver(None, False)
        self.assertRaises(NotImplementedError,
                          share_driver.update_replicated_snapshot,
                          'fake_context', ['r1', 'r2'], 'r1',
                          ['s1', 's2'], 's1')

    @ddt.data(True, False)
    def test_share_group_snapshot_support_exists_and_equals_snapshot_support(
            self, snapshots_are_supported):
        driver.CONF.set_default('driver_handles_share_servers', True)
        child_class_instance = driver.ShareDriver(True)
        child_class_instance._snapshots_are_supported = snapshots_are_supported
        self.mock_object(child_class_instance, "configuration")

        child_class_instance._update_share_stats()

        self.assertEqual(
            snapshots_are_supported,
            child_class_instance._stats["snapshot_support"])
        self.assertTrue(child_class_instance.configuration.safe_get.called)

    def test_create_share_group_from_share_group_snapshot(self):
        share_driver = self._instantiate_share_driver(None, False)
        fake_shares = [
            {'id': 'fake_share_%d' % i,
             'source_share_group_snapshot_member_id': 'fake_member_%d' % i}
            for i in (1, 2)]
        fake_share_group_dict = {
            'source_share_group_snapshot_id': 'some_fake_uuid_abc',
            'shares': fake_shares,
            'id': 'some_fake_uuid_def',
        }
        fake_share_group_snapshot_dict = {
            'share_group_snapshot_members': [
                {'id': 'fake_member_1'}, {'id': 'fake_member_2'}],
            'id': 'fake_share_group_snapshot_id',
        }
        mock_create = self.mock_object(
            share_driver, 'create_share_from_snapshot',
            mock.Mock(side_effect=['fake_export1', 'fake_export2']))
        expected_share_updates = [
            {
                'id': 'fake_share_1',
                'export_locations': 'fake_export1',
            },
            {
                'id': 'fake_share_2',
                'export_locations': 'fake_export2',
            },
        ]

        share_group_update, share_update = (
            share_driver.create_share_group_from_share_group_snapshot(
                'fake_context', fake_share_group_dict,
                fake_share_group_snapshot_dict))

        mock_create.assert_has_calls([
            mock.call(
                'fake_context',
                {'id': 'fake_share_1',
                 'source_share_group_snapshot_member_id': 'fake_member_1'},
                {'id': 'fake_member_1'}),
            mock.call(
                'fake_context',
                {'id': 'fake_share_2',
                 'source_share_group_snapshot_member_id': 'fake_member_2'},
                {'id': 'fake_member_2'})
        ])
        self.assertIsNone(share_group_update)
        self.assertEqual(expected_share_updates, share_update)

    def test_create_share_group_from_share_group_snapshot_dhss(self):
        share_driver = self._instantiate_share_driver(None, True)
        mock_share_server = mock.Mock()
        fake_shares = [
            {'id': 'fake_share_1',
             'source_share_group_snapshot_member_id': 'foo_member_1'},
            {'id': 'fake_share_2',
             'source_share_group_snapshot_member_id': 'foo_member_2'}]
        fake_share_group_dict = {
            'source_share_group_snapshot_id': 'some_fake_uuid',
            'shares': fake_shares,
            'id': 'eda52174-0442-476d-9694-a58327466c14',
        }
        fake_share_group_snapshot_dict = {
            'share_group_snapshot_members': [
                {'id': 'foo_member_1'}, {'id': 'foo_member_2'}],
            'id': 'fake_share_group_snapshot_id'
        }
        mock_create = self.mock_object(
            share_driver, 'create_share_from_snapshot',
            mock.Mock(side_effect=['fake_export1', 'fake_export2']))
        expected_share_updates = [
            {'id': 'fake_share_1', 'export_locations': 'fake_export1'},
            {'id': 'fake_share_2', 'export_locations': 'fake_export2'},
        ]

        share_group_update, share_update = (
            share_driver.create_share_group_from_share_group_snapshot(
                'fake_context',
                fake_share_group_dict,
                fake_share_group_snapshot_dict, share_server=mock_share_server,
            )
        )

        mock_create.assert_has_calls([
            mock.call(
                'fake_context',
                {'id': 'fake_share_%d' % i,
                 'source_share_group_snapshot_member_id': 'foo_member_%d' % i},
                {'id': 'foo_member_%d' % i},
                share_server=mock_share_server)
            for i in (1, 2)
        ])
        self.assertIsNone(share_group_update)
        self.assertEqual(expected_share_updates, share_update)

    def test_create_share_group_from_share_group_snapshot_with_dict_raise(
            self):
        share_driver = self._instantiate_share_driver(None, False)
        fake_shares = [
            {'id': 'fake_share_%d' % i,
             'source_share_group_snapshot_member_id': 'fake_member_%d' % i}
            for i in (1, 2)]
        fake_share_group_dict = {
            'source_share_group_snapshot_id': 'some_fake_uuid_abc',
            'shares': fake_shares,
            'id': 'some_fake_uuid_def',
        }
        fake_share_group_snapshot_dict = {
            'share_group_snapshot_members': [
                {'id': 'fake_member_1'}, {'id': 'fake_member_2'}],
            'id': 'fake_share_group_snapshot_id',
        }

        self.mock_object(
            share_driver, 'create_share_from_snapshot',
            mock.Mock(side_effect=[{
                'export_locations': 'fake_export1',
                'status': constants.STATUS_CREATING},
                {'export_locations': 'fake_export2',
                 'status': constants.STATUS_CREATING}]))

        self.assertRaises(
            exception.InvalidShareInstance,
            share_driver.create_share_group_from_share_group_snapshot,
            'fake_context',
            fake_share_group_dict,
            fake_share_group_snapshot_dict)

    def test_create_share_group_from_share_group_snapshot_with_dict(
            self):
        share_driver = self._instantiate_share_driver(None, False)
        fake_shares = [
            {'id': 'fake_share_%d' % i,
             'source_share_group_snapshot_member_id': 'fake_member_%d' % i}
            for i in (1, 2)]
        fake_share_group_dict = {
            'source_share_group_snapshot_id': 'some_fake_uuid_abc',
            'shares': fake_shares,
            'id': 'some_fake_uuid_def',
        }
        fake_share_group_snapshot_dict = {
            'share_group_snapshot_members': [
                {'id': 'fake_member_1'}, {'id': 'fake_member_2'}],
            'id': 'fake_share_group_snapshot_id',
        }

        mock_create = self.mock_object(
            share_driver, 'create_share_from_snapshot',
            mock.Mock(side_effect=[{
                'export_locations': 'fake_export1',
                'status': constants.STATUS_CREATING_FROM_SNAPSHOT},
                {'export_locations': 'fake_export2',
                 'status': constants.STATUS_AVAILABLE}]))
        expected_share_updates = [
            {
                'id': 'fake_share_1',
                'status': constants.STATUS_CREATING_FROM_SNAPSHOT,
                'export_locations': 'fake_export1',
            },
            {
                'id': 'fake_share_2',
                'status': constants.STATUS_AVAILABLE,
                'export_locations': 'fake_export2',
            },
        ]

        share_group_update, share_update = (
            share_driver.create_share_group_from_share_group_snapshot(
                'fake_context', fake_share_group_dict,
                fake_share_group_snapshot_dict))

        mock_create.assert_has_calls([
            mock.call(
                'fake_context',
                {'id': 'fake_share_1',
                 'source_share_group_snapshot_member_id': 'fake_member_1'},
                {'id': 'fake_member_1'}),
            mock.call(
                'fake_context',
                {'id': 'fake_share_2',
                 'source_share_group_snapshot_member_id': 'fake_member_2'},
                {'id': 'fake_member_2'})
        ])
        self.assertIsNone(share_group_update)
        self.assertEqual(expected_share_updates, share_update)

    def test_update_share_server_security_service(self):
        share_driver = self._instantiate_share_driver(None, True)
        self.assertRaises(NotImplementedError,
                          share_driver.update_share_server_security_service,
                          'fake_context',
                          {'id', 'share_server_id'},
                          {'fake', 'fake_net_info'},
                          [{"id": "fake_instance_id"}],
                          [{"id": "fake_rule_id"}],
                          {'id', 'fake_sec_service_id'},
                          current_security_service=None)

    def test_check_update_share_server_security_service(self):
        share_driver = self._instantiate_share_driver(None, True)
        self.assertRaises(
            NotImplementedError,
            share_driver.check_update_share_server_security_service,
            'fake_context',
            {'id', 'share_server_id'},
            {'fake', 'fake_net_info'},
            [{"id": "fake_instance_id"}],
            [{"id": "fake_rule_id"}],
            {'id', 'fake_sec_service_id'},
            current_security_service=None)

    def test_create_share_group_from_sg_snapshot_with_no_members(self):
        share_driver = self._instantiate_share_driver(None, False)
        fake_share_group_dict = {}
        fake_share_group_snapshot_dict = {'share_group_snapshot_members': []}

        share_group_update, share_update = (
            share_driver.create_share_group_from_share_group_snapshot(
                'fake_context', fake_share_group_dict,
                fake_share_group_snapshot_dict))

        self.assertIsNone(share_group_update)
        self.assertIsNone(share_update)

    def test_create_share_group_snapshot(self):
        fake_snap_member_1 = {
            'id': '6813e06b-a8f5-4784-b17d-f3e91afa370e',
            'share_id': 'a3ebdba5-b4e1-46c8-a0ea-a9ac8daf5296',
            'share_group_snapshot_id': 'fake_share_group_snapshot_id',
            'share_instance_id': 'fake_share_instance_id_1',
            'provider_location': 'should_not_be_used_1',
            'share_name': 'share_fake_share_instance_id_1',
            'name': 'share-snapshot-6813e06b-a8f5-4784-b17d-f3e91afa370e',
            'share': {
                'id': '420f978b-dbf6-4b3c-92fe-f5b17a0bb5e2',
                'size': 3,
                'share_proto': 'fake_share_proto',
            },
        }
        fake_snap_member_2 = {
            'id': '1e010dfe-545b-432d-ab95-4ef03cd82f89',
            'share_id': 'a3ebdba5-b4e1-46c8-a0ea-a9ac8daf5296',
            'share_group_snapshot_id': 'fake_share_group_snapshot_id',
            'share_instance_id': 'fake_share_instance_id_2',
            'provider_location': 'should_not_be_used_2',
            'share_name': 'share_fake_share_instance_id_2',
            'name': 'share-snapshot-1e010dfe-545b-432d-ab95-4ef03cd82f89',
            'share': {
                'id': '420f978b-dbf6-4b3c-92fe-f5b17a0bb5e2',
                'size': '2',
                'share_proto': 'fake_share_proto',
            },
        }
        fake_snap_dict = {
            'status': 'available',
            'project_id': '13c0be6290934bd98596cfa004650049',
            'user_id': 'a0314a441ca842019b0952224aa39192',
            'description': None,
            'deleted': '0',
            'share_group_id': '4b04fdc3-00b9-4909-ba1a-06e9b3f88b67',
            'share_group_snapshot_members': [
                fake_snap_member_1, fake_snap_member_2],
            'deleted_at': None,
            'id': 'f6aa3b59-57eb-421e-965c-4e182538e36a',
            'name': None
        }
        share_driver = self._instantiate_share_driver(None, False)
        share_driver._stats['snapshot_support'] = True
        mock_create_snap = self.mock_object(
            share_driver, 'create_snapshot',
            mock.Mock(side_effect=lambda *args, **kwargs: {
                'foo_k': 'foo_v', 'bar_k': 'bar_v_%s' % args[1]['id']}))

        share_group_snapshot_update, member_update_list = (
            share_driver.create_share_group_snapshot(
                'fake_context', fake_snap_dict))

        mock_create_snap.assert_has_calls([
            mock.call(
                'fake_context',
                {'snapshot_id': member['share_group_snapshot_id'],
                 'share_id': member['share_id'],
                 'share_instance_id': member['share']['id'],
                 'id': member['id'],
                 'share': member['share'],
                 'share_name': member['share_name'],
                 'name': member['name'],
                 'size': member['share']['size'],
                 'share_size': member['share']['size'],
                 'share_proto': member['share']['share_proto'],
                 'provider_location': None},
                share_server=None)
            for member in (fake_snap_member_1, fake_snap_member_2)
        ])
        self.assertIsNone(share_group_snapshot_update)
        self.assertEqual(
            [{'id': member['id'], 'foo_k': 'foo_v',
              'bar_k': 'bar_v_%s' % member['id']}
             for member in (fake_snap_member_1, fake_snap_member_2)],
            member_update_list,
        )

    def test_create_share_group_snapshot_failed_snapshot(self):
        fake_snap_member_1 = {
            'id': '6813e06b-a8f5-4784-b17d-f3e91afa370e',
            'share_id': 'a3ebdba5-b4e1-46c8-a0ea-a9ac8daf5296',
            'share_group_snapshot_id': 'fake_share_group_snapshot_id',
            'share_instance_id': 'fake_share_instance_id_1',
            'provider_location': 'should_not_be_used_1',
            'share_name': 'share_fake_share_instance_id_1',
            'name': 'share-snapshot-6813e06b-a8f5-4784-b17d-f3e91afa370e',
            'share': {
                'id': '420f978b-dbf6-4b3c-92fe-f5b17a0bb5e2',
                'size': 3,
                'share_proto': 'fake_share_proto',
            },
        }
        fake_snap_member_2 = {
            'id': '1e010dfe-545b-432d-ab95-4ef03cd82f89',
            'share_id': 'a3ebdba5-b4e1-46c8-a0ea-a9ac8daf5296',
            'share_group_snapshot_id': 'fake_share_group_snapshot_id',
            'share_instance_id': 'fake_share_instance_id_2',
            'provider_location': 'should_not_be_used_2',
            'share_name': 'share_fake_share_instance_id_2',
            'name': 'share-snapshot-1e010dfe-545b-432d-ab95-4ef03cd82f89',
            'share': {
                'id': '420f978b-dbf6-4b3c-92fe-f5b17a0bb5e2',
                'size': '2',
                'share_proto': 'fake_share_proto',
            },
        }
        fake_snap_dict = {
            'status': 'available',
            'project_id': '13c0be6290934bd98596cfa004650049',
            'user_id': 'a0314a441ca842019b0952224aa39192',
            'description': None,
            'deleted': '0',
            'share_group_id': '4b04fdc3-00b9-4909-ba1a-06e9b3f88b67',
            'share_group_snapshot_members': [
                fake_snap_member_1, fake_snap_member_2],
            'deleted_at': None,
            'id': 'f6aa3b59-57eb-421e-965c-4e182538e36a',
            'name': None
        }
        expected_exception = exception.ManilaException

        share_driver = self._instantiate_share_driver(None, False)
        share_driver._stats['snapshot_support'] = True
        mock_create_snap = self.mock_object(
            share_driver, 'create_snapshot',
            mock.Mock(side_effect=[None, expected_exception]))
        mock_delete_snap = self.mock_object(share_driver, 'delete_snapshot')

        self.assertRaises(
            expected_exception,
            share_driver.create_share_group_snapshot,
            'fake_context', fake_snap_dict)

        fake_snap_member_1_expected = {
            'snapshot_id': fake_snap_member_1['share_group_snapshot_id'],
            'share_id': fake_snap_member_1['share_id'],
            'share_instance_id': fake_snap_member_1['share']['id'],
            'id': fake_snap_member_1['id'],
            'share': fake_snap_member_1['share'],
            'share_name': fake_snap_member_1['share_name'],
            'name': fake_snap_member_1['name'],
            'size': fake_snap_member_1['share']['size'],
            'share_size': fake_snap_member_1['share']['size'],
            'share_proto': fake_snap_member_1['share']['share_proto'],
            'provider_location': None,
        }
        mock_create_snap.assert_has_calls([
            mock.call(
                'fake_context',
                {'snapshot_id': member['share_group_snapshot_id'],
                 'share_id': member['share_id'],
                 'share_instance_id': member['share']['id'],
                 'id': member['id'],
                 'share': member['share'],
                 'share_name': member['share_name'],
                 'name': member['name'],
                 'size': member['share']['size'],
                 'share_size': member['share']['size'],
                 'share_proto': member['share']['share_proto'],
                 'provider_location': None},
                share_server=None)
            for member in (fake_snap_member_1, fake_snap_member_2)
        ])
        mock_delete_snap.assert_called_with(
            'fake_context', fake_snap_member_1_expected, share_server=None)

    def test_create_share_group_snapshot_no_support(self):
        fake_snap_dict = {
            'status': 'available',
            'project_id': '13c0be6290934bd98596cfa004650049',
            'user_id': 'a0314a441ca842019b0952224aa39192',
            'description': None,
            'deleted': '0',
            'share_group_id': '4b04fdc3-00b9-4909-ba1a-06e9b3f88b67',
            'share_group_snapshot_members': [
                {
                    'status': 'available',
                    'share_type_id': '1a9ed31e-ee70-483d-93ba-89690e028d7f',
                    'user_id': 'a0314a441ca842019b0952224aa39192',
                    'deleted': 'False',
                    'share_proto': 'NFS',
                    'project_id': '13c0be6290934bd98596cfa004650049',
                    'share_group_snapshot_id':
                        'f6aa3b59-57eb-421e-965c-4e182538e36a',
                    'deleted_at': None,
                    'id': '6813e06b-a8f5-4784-b17d-f3e91afa370e',
                    'size': 1
                },
            ],
            'deleted_at': None,
            'id': 'f6aa3b59-57eb-421e-965c-4e182538e36a',
            'name': None
        }
        share_driver = self._instantiate_share_driver(None, False)
        share_driver._stats['snapshot_support'] = False

        self.assertRaises(
            exception.ShareGroupSnapshotNotSupported,
            share_driver.create_share_group_snapshot,
            'fake_context', fake_snap_dict)

    def test_create_share_group_snapshot_no_members(self):
        fake_snap_dict = {
            'status': 'available',
            'project_id': '13c0be6290934bd98596cfa004650049',
            'user_id': 'a0314a441ca842019b0952224aa39192',
            'description': None,
            'deleted': '0',
            'share_group_id': '4b04fdc3-00b9-4909-ba1a-06e9b3f88b67',
            'share_group_snapshot_members': [],
            'deleted_at': None,
            'id': 'f6aa3b59-57eb-421e-965c-4e182538e36a',
            'name': None
        }
        share_driver = self._instantiate_share_driver(None, False)
        share_driver._stats['snapshot_support'] = True

        share_group_snapshot_update, member_update_list = (
            share_driver.create_share_group_snapshot(
                'fake_context', fake_snap_dict))

        self.assertIsNone(share_group_snapshot_update)
        self.assertIsNone(member_update_list)

    def test_delete_share_group_snapshot(self):
        fake_snap_member_1 = {
            'id': '6813e06b-a8f5-4784-b17d-f3e91afa370e',
            'share_id': 'a3ebdba5-b4e1-46c8-a0ea-a9ac8daf5296',
            'share_group_snapshot_id': 'fake_share_group_snapshot_id',
            'share_instance_id': 'fake_share_instance_id_1',
            'provider_location': 'fake_provider_location_2',
            'share_name': 'share_fake_share_instance_id_1',
            'name': 'share-snapshot-6813e06b-a8f5-4784-b17d-f3e91afa370e',
            'share': {
                'id': '420f978b-dbf6-4b3c-92fe-f5b17a0bb5e2',
                'size': 3,
                'share_proto': 'fake_share_proto',
            },
        }
        fake_snap_member_2 = {
            'id': '1e010dfe-545b-432d-ab95-4ef03cd82f89',
            'share_id': 'a3ebdba5-b4e1-46c8-a0ea-a9ac8daf5296',
            'share_group_snapshot_id': 'fake_share_group_snapshot_id',
            'share_instance_id': 'fake_share_instance_id_2',
            'provider_location': 'fake_provider_location_2',
            'share_name': 'share_fake_provider_location_2',
            'name': 'share-snapshot-1e010dfe-545b-432d-ab95-4ef03cd82f89',
            'share': {
                'id': '420f978b-dbf6-4b3c-92fe-f5b17a0bb5e2',
                'size': '2',
                'share_proto': 'fake_share_proto',
            },
        }
        fake_snap_dict = {
            'status': 'available',
            'project_id': '13c0be6290934bd98596cfa004650049',
            'user_id': 'a0314a441ca842019b0952224aa39192',
            'description': None,
            'deleted': '0',
            'share_group_id': '4b04fdc3-00b9-4909-ba1a-06e9b3f88b67',
            'share_group_snapshot_members': [
                fake_snap_member_1, fake_snap_member_2],
            'deleted_at': None,
            'id': 'f6aa3b59-57eb-421e-965c-4e182538e36a',
            'name': None
        }

        share_driver = self._instantiate_share_driver(None, False)
        share_driver._stats['share_group_snapshot_support'] = True
        mock_delete_snap = self.mock_object(share_driver, 'delete_snapshot')

        share_group_snapshot_update, member_update_list = (
            share_driver.delete_share_group_snapshot(
                'fake_context', fake_snap_dict))

        mock_delete_snap.assert_has_calls([
            mock.call(
                'fake_context',
                {'snapshot_id': member['share_group_snapshot_id'],
                 'share_id': member['share_id'],
                 'share_instance_id': member['share']['id'],
                 'id': member['id'],
                 'share': member['share'],
                 'size': member['share']['size'],
                 'share_size': member['share']['size'],
                 'share_name': member['share_name'],
                 'name': member['name'],
                 'share_proto': member['share']['share_proto'],
                 'provider_location': member['provider_location']},
                share_server=None)
            for member in (fake_snap_member_1, fake_snap_member_2)
        ])
        self.assertIsNone(share_group_snapshot_update)
        self.assertIsNone(member_update_list)

    def test_snapshot_update_access(self):
        share_driver = self._instantiate_share_driver(None, False)
        self.assertRaises(NotImplementedError,
                          share_driver.snapshot_update_access,
                          'fake_context', 'fake_snapshot', ['r1', 'r2'],
                          [], [])

    @ddt.data({'user_networks': set([4]), 'conf': [4],
               'expected': {'ipv4': True, 'ipv6': False}},
              {'user_networks': set([6]), 'conf': [4],
               'expected': {'ipv4': False, 'ipv6': False}},
              {'user_networks': set([4, 6]), 'conf': [4],
               'expected': {'ipv4': True, 'ipv6': False}},
              {'user_networks': set([4]), 'conf': [6],
               'expected': {'ipv4': False, 'ipv6': False}},
              {'user_networks': set([6]), 'conf': [6],
               'expected': {'ipv4': False, 'ipv6': True}},
              {'user_networks': set([4, 6]), 'conf': [6],
               'expected': {'ipv4': False, 'ipv6': True}},
              {'user_networks': set([4]), 'conf': [4, 6],
               'expected': {'ipv4': True, 'ipv6': False}},
              {'user_networks': set([6]), 'conf': [4, 6],
               'expected': {'ipv4': False, 'ipv6': True}},
              {'user_networks': set([4, 6]), 'conf': [4, 6],
               'expected': {'ipv4': True, 'ipv6': True}},
              )
    @ddt.unpack
    def test_add_ip_version_capability_if_dhss_true(self,
                                                    user_networks,
                                                    conf,
                                                    expected):
        share_driver = self._instantiate_share_driver(None, True)
        self.mock_object(share_driver, 'get_configured_ip_versions',
                         mock.Mock(return_value=conf))
        versions = mock.PropertyMock(return_value=user_networks)
        type(share_driver.network_api).enabled_ip_versions = versions
        data = {'share_backend_name': 'fake_backend'}

        result = share_driver.add_ip_version_capability(data)

        self.assertIsNotNone(result['ipv4_support'])
        self.assertEqual(expected['ipv4'], result['ipv4_support'])
        self.assertIsNotNone(result['ipv6_support'])
        self.assertEqual(expected['ipv6'], result['ipv6_support'])

    @ddt.data({'conf': [4],
               'expected': {'ipv4': True, 'ipv6': False}},
              {'conf': [6],
               'expected': {'ipv4': False, 'ipv6': True}},
              {'conf': [4, 6],
               'expected': {'ipv4': True, 'ipv6': True}},
              )
    @ddt.unpack
    def test_add_ip_version_capability_if_dhss_false(self, conf, expected):
        share_driver = self._instantiate_share_driver(None, False)
        self.mock_object(share_driver, 'get_configured_ip_versions',
                         mock.Mock(return_value=conf))
        data = {'share_backend_name': 'fake_backend'}
        result = share_driver.add_ip_version_capability(data)

        self.assertIsNotNone(result['ipv4_support'])
        self.assertEqual(expected['ipv4'], result['ipv4_support'])
        self.assertIsNotNone(result['ipv6_support'])
        self.assertEqual(expected['ipv6'], result['ipv6_support'])



# Copyright (c) 2016 QNAP Systems, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

FAKE_RES_DETAIL_DATA_LOGIN = """
    <QDocRoot version="1.0">
        <authPassed><![CDATA[1]]></authPassed>
        <authSid><![CDATA[fakeSid]]></authSid>
    </QDocRoot>"""

FAKE_RES_DETAIL_DATA_GETBASIC_INFO_ES_1_1_1 = """
    <QDocRoot version="1.0">
        <model>
            <displayModelName><![CDATA[ES1640dc]]></displayModelName>
            <internalModelName><![CDATA[ES1640dc]]></internalModelName>
        </model>
        <firmware>
            <version><![CDATA[1.1.1]]></version>
        </firmware>
    </QDocRoot>"""

FAKE_RES_DETAIL_DATA_GETBASIC_INFO_ES_1_1_3 = """
    <QDocRoot version="1.0">
        <model>
            <displayModelName><![CDATA[ES1640dc]]></displayModelName>
            <internalModelName><![CDATA[ES1640dc]]></internalModelName>
        </model>
        <firmware>
            <version><![CDATA[1.1.3]]></version>
        </firmware>
    </QDocRoot>"""

FAKE_RES_DETAIL_DATA_GETBASIC_INFO_ES_2_0_0 = """
    <QDocRoot version="1.0">
        <model>
            <displayModelName><![CDATA[ES1640dc]]></displayModelName>
            <internalModelName><![CDATA[ES1640dc]]></internalModelName>
        </model>
        <firmware>
            <version><![CDATA[2.0.0]]></version>
        </firmware>
    </QDocRoot>"""

FAKE_RES_DETAIL_DATA_GETBASIC_INFO_ES_2_1_0 = """
    <QDocRoot version="1.0">
        <model>
            <displayModelName><![CDATA[ES1640dc]]></displayModelName>
            <internalModelName><![CDATA[ES1640dc]]></internalModelName>
        </model>
        <firmware>
            <version><![CDATA[2.1.0]]></version>
        </firmware>
    </QDocRoot>"""

FAKE_RES_DETAIL_DATA_GETBASIC_INFO_ES_2_2_0 = """
    <QDocRoot version="1.0">
        <model>
            <displayModelName><![CDATA[ES1640dc]]></displayModelName>
            <internalModelName><![CDATA[ES1640dc]]></internalModelName>
        </model>
        <firmware>
            <version><![CDATA[2.2.0]]></version>
        </firmware>
    </QDocRoot>"""

FAKE_RES_DETAIL_DATA_GETBASIC_INFO_TS_4_0_0 = """
    <QDocRoot version="1.0">
        <model>
            <displayModelName><![CDATA[TS-870U]]></displayModelName>
            <internalModelName><![CDATA[TS-870U]]></internalModelName>
        </model>
        <firmware>
            <version><![CDATA[4.0.0]]></version>
        </firmware>
    </QDocRoot>"""

FAKE_RES_DETAIL_DATA_GETBASIC_INFO_TS_4_3_0 = """
    <QDocRoot version="1.0">
        <model>
            <displayModelName><![CDATA[TS-870U]]></displayModelName>
            <internalModelName><![CDATA[TS-870U]]></internalModelName>
        </model>
        <firmware>
            <version><![CDATA[4.3.0]]></version>
        </firmware>
    </QDocRoot>"""

FAKE_RES_DETAIL_DATA_GETBASIC_INFO_TES_TS_4_0_0 = """
    <QDocRoot version="1.0">
        <model>
            <displayModelName><![CDATA[TES-1885U]]></displayModelName>
            <internalModelName><![CDATA[TS-1885U]]></internalModelName>
        </model>
        <firmware>
            <version><![CDATA[4.0.0]]></version>
        </firmware>
    </QDocRoot>"""

FAKE_RES_DETAIL_DATA_GETBASIC_INFO_TES_TS_4_3_0 = """
    <QDocRoot version="1.0">
        <model>
            <displayModelName><![CDATA[TES-1885U]]></displayModelName>
            <internalModelName><![CDATA[TS-1885U]]></internalModelName>
        </model>
        <firmware>
            <version><![CDATA[4.3.0]]></version>
        </firmware>
    </QDocRoot>"""

FAKE_RES_DETAIL_DATA_GETBASIC_INFO_TES_ES_1_1_1 = """
    <QDocRoot version="1.0">
        <model>
            <displayModelName><![CDATA[TES-1885U]]></displayModelName>
            <internalModelName><![CDATA[ES-1885U]]></internalModelName>
        </model>
        <firmware>
            <version><![CDATA[1.1.1]]></version>
        </firmware>
    </QDocRoot>"""

FAKE_RES_DETAIL_DATA_GETBASIC_INFO_TES_ES_1_1_3 = """
    <QDocRoot version="1.0">
        <model>
            <displayModelName><![CDATA[TES-1885U]]></displayModelName>
            <internalModelName><![CDATA[ES-1885U]]></internalModelName>
        </model>
        <firmware>
            <version><![CDATA[1.1.3]]></version>
        </firmware>
    </QDocRoot>"""

FAKE_RES_DETAIL_DATA_GETBASIC_INFO_TES_ES_2_0_0 = """
    <QDocRoot version="1.0">
        <model>
            <displayModelName><![CDATA[TES-1885U]]></displayModelName>
            <internalModelName><![CDATA[ES-1885U]]></internalModelName>
        </model>
        <firmware>
            <version><![CDATA[2.0.0]]></version>
        </firmware>
    </QDocRoot>"""

FAKE_RES_DETAIL_DATA_GETBASIC_INFO_TES_ES_2_1_0 = """
    <QDocRoot version="1.0">
        <model>
            <displayModelName><![CDATA[TES-1885U]]></displayModelName>
            <internalModelName><![CDATA[ES-1885U]]></internalModelName>
        </model>
        <firmware>
            <version><![CDATA[2.1.0]]></version>
        </firmware>
    </QDocRoot>"""

FAKE_RES_DETAIL_DATA_GETBASIC_INFO_TES_ES_2_2_0 = """
    <QDocRoot version="1.0">
        <model>
            <displayModelName><![CDATA[TES-1885U]]></displayModelName>
            <internalModelName><![CDATA[ES-1885U]]></internalModelName>
        </model>
        <firmware>
            <version><![CDATA[2.2.0]]></version>
        </firmware>
    </QDocRoot>"""

FAKE_RES_DETAIL_DATA_GETBASIC_INFO_ERROR = """
    <QDocRoot version="1.0">
        <model>
            <displayModelName><![CDATA[TTS-1885U]]></displayModelName>
            <internalModelName><![CDATA[TTS-1885U]]></internalModelName>
        </model>
        <firmware>
            <version><![CDATA[1.1.3]]></version>
        </firmware>
    </QDocRoot>"""

FAKE_RES_DETAIL_DATA_SHARE_INFO = """
    <QDocRoot version="1.0">
        <authPassed><![CDATA[1]]></authPassed>
        <Volume_Info>
            <row>
                <vol_no><![CDATA[fakeNo]]></vol_no>
                <vol_label><![CDATA[fakeShareName]]></vol_label>
            </row>
        </Volume_Info>
        <result><![CDATA[0]]></result>
    </QDocRoot>"""

FAKE_RES_DETAIL_DATA_VOLUME_INFO = """
    <QDocRoot version="1.0">
        <authPassed><![CDATA[1]]></authPassed>
        <Volume_Info>
            <row>
                <vol_no><![CDATA[fakeNo]]></vol_no>
                <size><![CDATA[10]]></size>
                <vol_mount_path>fakeMountPath</vol_mount_path>
                <dedup><![CDATA[off]]></dedup>
                <compression><![CDATA[1]]></compression>
                <thin_pro><![CDATA[1]]></thin_pro>
                <cache><![CDATA[0]]></cache>
            </row>
        </Volume_Info>
        <result><![CDATA[0]]></result>
    </QDocRoot>"""

FAKE_RES_DETAIL_DATA_SNAPSHOT = """
    <QDocRoot version="1.0">
        <authPassed><![CDATA[1]]></authPassed>
        <SnapshotList>
            <row>
                <snapshot_id><![CDATA[fakeSnapshotId]]></snapshot_id>
                <snapshot_name><![CDATA[fakeSnapshotName]]></snapshot_name>
                <parent_size>10</parent_size>
            </row>
        </SnapshotList>
        <result><![CDATA[0]]></result>
    </QDocRoot>"""

FAKE_RES_DETAIL_DATA_SPECIFIC_POOL_INFO = """
    <QDocRoot version="1.0">
        <authPassed><![CDATA[1]]></authPassed>
        <Pool_Index>
            <row>
                <poolIndex><![CDATA[fakePoolIndex]]></poolIndex>
                <poolID><![CDATA[fakePoolId]]></poolID>
                <pool_status><![CDATA[0]]></pool_status>
                <capacity_bytes><![CDATA[930213412209]]></capacity_bytes>
                <allocated_bytes><![CDATA[1480470528]]></allocated_bytes>
                <freesize_bytes><![CDATA[928732941681]]></freesize_bytes>
                <lun_meta_reserve_ratio><![CDATA[0.0315]]></lun_meta_reserve_ratio>
                <pool_capacity><![CDATA[866 GB]]></pool_capacity>
                <pool_allocated><![CDATA[1.38 GB]]></pool_allocated>
                <pool_freesize><![CDATA[865 GB]]></pool_freesize>
                <pool_threshold><![CDATA[80 %]]></pool_threshold>
                <pool_used><![CDATA[0 %]]></pool_used>
                <pool_available><![CDATA[100 %]]></pool_available>
                <pool_owner><![CDATA[SCA]]></pool_owner>
                <pool_type><![CDATA[mirror]]></pool_type>
                <pool_dedup><![CDATA[1.00]]></pool_dedup>
                <pool_bound><![CDATA[0]]></pool_bound>
                <pool_progress><![CDATA[0]]></pool_progress>
                <pool_scrub><![CDATA[0]]></pool_scrub>
            </row>
        </Pool_Index>
        <result><![CDATA[0]]></result>
    </QDocRoot>"""

FAKE_RES_DETAIL_DATA_GET_HOST_LIST = """
    <QDocRoot version="1.0">
        <authPassed><![CDATA[1]]></authPassed>
        <ES_RET_CODE><![CDATA[1]]></ES_RET_CODE>
        <host_list>
            <host>
                <index><![CDATA[fakeHostIndex]]></index>
                <hostid><![CDATA[fakeHostId]]></hostid>
                <name><![CDATA[manila-fakeHostName]]></name>
                <netaddrs>
                    <ipv4>
                        <![CDATA[fakeIp]]>
                    </ipv4>
                </netaddrs>
            </host>
        </host_list>
        <result><![CDATA[0]]></result>
    </QDocRoot>"""

FAKE_RES_DETAIL_DATA_CREATE_SHARE = """
    <QDocRoot version="1.0">
        <authPassed><![CDATA[1]]></authPassed>
        <ES_RET_CODE><![CDATA[1]]></ES_RET_CODE>
        <func>
            <ownContent>
                <volumeList>
                    <volume>
                        <volumeStatus><![CDATA[fakeStatus]]></volumeStatus>
                        <volumeLabel><![CDATA[fakeLabel]]></volumeLabel>
                        <volumeValue><![CDATA[faleValue]]></volumeValue>
                    </volume>
                </volumeList>
            </ownContent>
        </func>
    </QDocRoot>"""

FAKE_RES_DETAIL_DATA_ES_RET_CODE_NEGATIVE = """
    <QDocRoot version="1.0">
        <authPassed><![CDATA[1]]></authPassed>
        <ES_RET_CODE><![CDATA[-1]]></ES_RET_CODE>
    </QDocRoot>"""


FAKE_RES_DETAIL_DATA_RESULT_NEGATIVE = """
    <QDocRoot version="1.0">
        <authPassed><![CDATA[1]]></authPassed>
        <result><![CDATA[-1]]></result>
    </QDocRoot>"""

FAKE_RES_DETAIL_DATA_AUTHPASS_FAIL = """
    <QDocRoot version="1.0">
        <authPassed><![CDATA[0]]></authPassed>
    </QDocRoot>"""

FAKE_RES_DETAIL_DATA_DELETE_SHARE = """
    <QDocRoot version="1.0">
        <authPassed><![CDATA[1]]></authPassed>
        <ES_RET_CODE><![CDATA[1]]></ES_RET_CODE>
        <result>0</result>
    </QDocRoot>"""

FAKE_RES_DETAIL_DATA_DELETE_SNAPSHOT = """
    <QDocRoot version="1.0">
        <authPassed><![CDATA[1]]></authPassed>
        <ES_RET_CODE><![CDATA[1]]></ES_RET_CODE>
        <result>0</result>
    </QDocRoot>"""

FAKE_RES_DETAIL_DATA_DELETE_SNAPSHOT_SNAPSHOT_NOT_EXIST = """
    <QDocRoot version="1.0">
        <authPassed><![CDATA[1]]></authPassed>
        <ES_RET_CODE><![CDATA[1]]></ES_RET_CODE>
        <result>-206021</result>
    </QDocRoot>"""

FAKE_RES_DETAIL_DATA_DELETE_SNAPSHOT_SHARE_NOT_EXIST = """
    <QDocRoot version="1.0">
        <authPassed><![CDATA[1]]></authPassed>
        <ES_RET_CODE><![CDATA[1]]></ES_RET_CODE>
        <result>-200005</result>
    </QDocRoot>"""

FAKE_RES_DETAIL_DATA_GET_HOST_LIST_API = """
    <QDocRoot version="1.0">
        <authPassed><![CDATA[1]]></authPassed>
        <ES_RET_CODE><![CDATA[1]]></ES_RET_CODE>
        <content>
            <host_list>
                <host>
                    <index><![CDATA[fakeHostIndex]]></index>
                    <hostid><![CDATA[fakeHostId]]></hostid>
                    <name><![CDATA[manila-hst-123]]></name>
                    <netaddrs>
                        <ipv4>
                            <![CDATA[fakeIp]]>
                        </ipv4>
                    </netaddrs>
                </host>
            </host_list>
        </content>
        <result><![CDATA[0]]></result>
    </QDocRoot>"""

FAKE_RES_DETAIL_DATA_GET_NO_HOST_LIST_API = """
    <QDocRoot version="1.0">
        <authPassed><![CDATA[1]]></authPassed>
        <ES_RET_CODE><![CDATA[1]]></ES_RET_CODE>
        <content>
        </content>
        <result><![CDATA[0]]></result>
    </QDocRoot>"""

FAKE_RES_DETAIL_DATA_CREATE_SNAPSHOT = """
    <QDocRoot version="1.0">
        <authPassed><![CDATA[1]]></authPassed>
        <ES_RET_CODE><![CDATA[1]]></ES_RET_CODE>
    </QDocRoot>"""


class SnapshotClass(object):
    """Snapshot Class."""

    size = 0
    provider_location = 'fakeShareName@fakeSnapshotName'

    def __init__(self, size, provider_location=None):
        """Init."""
        self.size = size
        self.provider_location = provider_location

    def get(self, provider_location):
        """Get function."""
        return self.provider_location

    def __getitem__(self, arg):
        """Getitem."""
        return {
            'display_name': 'fakeSnapshotDisplayName',
            'id': 'fakeSnapshotId',
            'share': {'share_id': 'fakeShareId', 'id': 'fakeId'},
            'share_instance': {'share_id': 'fakeShareId', 'id': 'fakeId'},
            'size': self.size,
            'share_instance_id': 'fakeShareId'
        }[arg]

    def __setitem__(self, key, value):
        """Setitem."""
        if key == 'provider_location':
            self.provider_location = value


class ShareNfsClass(object):
    """Share Class."""

    share_proto = 'NFS'
    id = ''
    size = 0

    def __init__(self, share_id, size):
        """Init."""
        self.id = share_id
        self.size = size

    def __getitem__(self, arg):
        """Getitem."""
        return {
            'share_proto': self.share_proto,
            'id': self.id,
            'display_name': 'fakeDisplayName',
            'export_locations': [{'path': '1.2.3.4:/share/fakeShareName'}],
            'host': 'QnapShareDriver',
            'size': self.size
        }[arg]

    def __setitem__(self, key, value):
        """Setitem."""
        if key == 'share_proto':
            self.share_proto = value


class ShareCifsClass(object):
    """Share Class."""

    share_proto = 'CIFS'
    id = ''
    size = 0

    def __init__(self, share_id, size):
        """Init."""
        self.id = share_id
        self.size = size

    def __getitem__(self, arg):
        """Getitem."""
        return {
            'share_proto': self.share_proto,
            'id': self.id,
            'display_name': 'fakeDisplayName',
            'export_locations': [{'path': '\\\\1.2.3.4\\fakeShareName'}],
            'host': 'QnapShareDriver',
            'size': self.size
        }[arg]

    def __setitem__(self, key, value):
        """Setitem."""
        if key == 'share_proto':
            self.share_proto = value


class AccessClass(object):
    """Access Class."""

    access_type = 'fakeAccessType'
    access_level = 'ro'
    access_to = 'fakeIp'

    def __init__(self, access_type, access_level, access_to):
        """Init."""
        self.access_type = access_type
        self.access_level = access_level
        self.access_to = access_to

    def __getitem__(self, arg):
        """Getitem."""
        return {
            'access_type': self.access_type,
            'access_level': self.access_level,
            'access_to': self.access_to,
        }[arg]


class FakeGetBasicInfoResponseEs_1_1_1(object):
    """Fake GetBasicInfo response from ES nas."""

    status = 'fackStatus'

    def read(self):
        """Mock response.read."""
        return FAKE_RES_DETAIL_DATA_GETBASIC_INFO_ES_1_1_1


class FakeGetBasicInfoResponseEs_1_1_3(object):
    """Fake GetBasicInfo response from ES nas."""

    status = 'fackStatus'

    def read(self):
        """Mock response.read."""
        return FAKE_RES_DETAIL_DATA_GETBASIC_INFO_ES_1_1_3


class FakeGetBasicInfoResponseEs_2_0_0(object):
    """Fake GetBasicInfo response from ES nas."""

    status = 'fackStatus'

    def read(self):
        """Mock response.read."""
        return FAKE_RES_DETAIL_DATA_GETBASIC_INFO_ES_2_0_0


class FakeGetBasicInfoResponseEs_2_1_0(object):
    """Fake GetBasicInfo response from ES nas."""

    status = 'fackStatus'

    def read(self):
        """Mock response.read."""
        return FAKE_RES_DETAIL_DATA_GETBASIC_INFO_ES_2_1_0


class FakeGetBasicInfoResponseEs_2_2_0(object):
    """Fake GetBasicInfo response from ES nas."""

    status = 'fackStatus'

    def read(self):
        """Mock response.read."""
        return FAKE_RES_DETAIL_DATA_GETBASIC_INFO_ES_2_2_0


class FakeGetBasicInfoResponseTs_4_0_0(object):
    """Fake GetBasicInfoTS response from TS nas."""

    status = 'fackStatus'

    def read(self):
        """Mock response.read."""
        return FAKE_RES_DETAIL_DATA_GETBASIC_INFO_TS_4_0_0


class FakeGetBasicInfoResponseTs_4_3_0(object):
    """Fake GetBasicInfoTS response from TS nas."""

    status = 'fackStatus'

    def read(self):
        """Mock response.read."""
        return FAKE_RES_DETAIL_DATA_GETBASIC_INFO_TS_4_3_0


class FakeGetBasicInfoResponseTesTs_4_0_0(object):
    """Fake GetBasicInfoTS response from TS nas."""

    status = 'fackStatus'

    def read(self):
        """Mock response.read."""
        return FAKE_RES_DETAIL_DATA_GETBASIC_INFO_TES_TS_4_0_0


class FakeGetBasicInfoResponseTesTs_4_3_0(object):
    """Fake GetBasicInfoTS response from TS nas."""

    status = 'fackStatus'

    def read(self):
        """Mock response.read."""
        return FAKE_RES_DETAIL_DATA_GETBASIC_INFO_TES_TS_4_3_0


class FakeGetBasicInfoResponseTesEs_1_1_1(object):
    """Fake GetBasicInfoTS response from TS nas."""

    status = 'fackStatus'

    def read(self):
        """Mock response.read."""
        return FAKE_RES_DETAIL_DATA_GETBASIC_INFO_TES_ES_1_1_1


class FakeGetBasicInfoResponseTesEs_1_1_3(object):
    """Fake GetBasicInfoTS response from TS nas."""

    status = 'fackStatus'

    def read(self):
        """Mock response.read."""
        return FAKE_RES_DETAIL_DATA_GETBASIC_INFO_TES_ES_1_1_3


class FakeGetBasicInfoResponseTesEs_2_0_0(object):
    """Fake GetBasicInfoTS response from TS nas."""

    status = 'fackStatus'

    def read(self):
        """Mock response.read."""
        return FAKE_RES_DETAIL_DATA_GETBASIC_INFO_TES_ES_2_0_0


class FakeGetBasicInfoResponseTesEs_2_1_0(object):
    """Fake GetBasicInfoTS response from TS nas."""

    status = 'fackStatus'

    def read(self):
        """Mock response.read."""
        return FAKE_RES_DETAIL_DATA_GETBASIC_INFO_TES_ES_2_1_0


class FakeGetBasicInfoResponseTesEs_2_2_0(object):
    """Fake GetBasicInfoTS response from TS nas."""

    status = 'fackStatus'

    def read(self):
        """Mock response.read."""
        return FAKE_RES_DETAIL_DATA_GETBASIC_INFO_TES_ES_2_2_0


class FakeGetBasicInfoResponseError(object):
    """Fake GetBasicInfoTS response from TS nas."""

    status = 'fackStatus'

    def read(self):
        """Mock response.read."""
        return FAKE_RES_DETAIL_DATA_GETBASIC_INFO_ERROR


class FakeCreateShareResponse(object):
    """Fake login response."""

    status = 'fackStatus'

    def read(self):
        """Mock response.read."""
        return FAKE_RES_DETAIL_DATA_CREATE_SHARE


class FakeDeleteShareResponse(object):
    """Fake login response."""

    status = 'fackStatus'

    def read(self):
        """Mock response.read."""
        return FAKE_RES_DETAIL_DATA_DELETE_SHARE


class FakeDeleteSnapshotResponse(object):
    """Fake delete snapshot response."""

    status = 'fackStatus'

    def read(self):
        """Mock response.read."""
        return FAKE_RES_DETAIL_DATA_DELETE_SNAPSHOT


class FakeDeleteSnapshotResponseSnapshotNotExist(object):
    """Fake delete snapshot response."""

    status = 'fackStatus'

    def read(self):
        """Mock response.read."""
        return FAKE_RES_DETAIL_DATA_DELETE_SNAPSHOT_SNAPSHOT_NOT_EXIST


class FakeDeleteSnapshotResponseShareNotExist(object):
    """Fake delete snapshot response."""

    status = 'fackStatus'

    def read(self):
        """Mock response.read."""
        return FAKE_RES_DETAIL_DATA_DELETE_SNAPSHOT_SHARE_NOT_EXIST


class FakeGetHostListResponse(object):
    """Fake host info response."""

    status = 'fackStatus'

    def read(self):
        """Mock response.read."""
        return FAKE_RES_DETAIL_DATA_GET_HOST_LIST_API


class FakeGetNoHostListResponse(object):
    """Fake host info response."""

    status = 'fackStatus'

    def read(self):
        """Mock response.read."""
        return FAKE_RES_DETAIL_DATA_GET_NO_HOST_LIST_API


class FakeAuthPassFailResponse(object):
    """Fake pool info response."""

    status = 'fackStatus'

    def read(self):
        """Mock response.read."""
        return FAKE_RES_DETAIL_DATA_AUTHPASS_FAIL


class FakeEsResCodeNegativeResponse(object):
    """Fake pool info response."""

    status = 'fackStatus'

    def read(self):
        """Mock response.read."""
        return FAKE_RES_DETAIL_DATA_ES_RET_CODE_NEGATIVE


class FakeResultNegativeResponse(object):
    """Fake pool info response."""

    status = 'fackStatus'

    def read(self):
        """Mock response.read."""
        return FAKE_RES_DETAIL_DATA_RESULT_NEGATIVE


class FakeLoginResponse(object):
    """Fake login response."""

    status = 'fackStatus'

    def read(self):
        """Mock response.read."""
        return FAKE_RES_DETAIL_DATA_LOGIN


class FakeSpecificPoolInfoResponse(object):
    """Fake pool info response."""

    status = 'fackStatus'

    def read(self):
        """Mock response.read."""
        return FAKE_RES_DETAIL_DATA_SPECIFIC_POOL_INFO


class FakeShareInfoResponse(object):
    """Fake pool info response."""

    status = 'fackStatus'

    def read(self):
        """Mock response.read."""
        return FAKE_RES_DETAIL_DATA_SHARE_INFO


class FakeSnapshotInfoResponse(object):
    """Fake pool info response."""

    status = 'fackStatus'

    def read(self):
        """Mock response.read."""
        return FAKE_RES_DETAIL_DATA_SNAPSHOT


class FakeSpecificVolInfoResponse(object):
    """Fake pool info response."""

    status = 'fackStatus'

    def read(self):
        """Mock response.read."""
        return FAKE_RES_DETAIL_DATA_VOLUME_INFO


class FakeCreateSnapshotResponse(object):
    """Fake pool info response."""

    status = 'fackStatus'

    def read(self):
        """Mock response.read."""
        return FAKE_RES_DETAIL_DATA_CREATE_SNAPSHOT



# Copyright 2010 United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# Copyright 2011 Justin Santa Barbara
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Utilities and helper functions."""

import contextlib
import functools
import inspect
import os
import pyclbr
import re
import shutil
import sys
import tempfile
import tenacity
import time

from eventlet import pools
import logging
import netaddr
from oslo_concurrency import lockutils
from oslo_concurrency import processutils
from oslo_config import cfg
from oslo_log import log
from oslo_utils import encodeutils
from oslo_utils import importutils
from oslo_utils import netutils
from oslo_utils import strutils
from oslo_utils import timeutils
import paramiko
import six
from webob import exc


from manila.common import constants
from manila.db import api as db_api
from manila import exception
from manila.i18n import _


CONF = cfg.CONF
LOG = log.getLogger(__name__)
if hasattr('CONF', 'debug') and CONF.debug:
    logging.getLogger("paramiko").setLevel(logging.DEBUG)

_ISO8601_TIME_FORMAT_SUBSECOND = '%Y-%m-%dT%H:%M:%S.%f'
_ISO8601_TIME_FORMAT = '%Y-%m-%dT%H:%M:%S'

synchronized = lockutils.synchronized_with_prefix('manila-')


def isotime(at=None, subsecond=False):
    """Stringify time in ISO 8601 format."""

    # Python provides a similar instance method for datetime.datetime objects
    # called isoformat(). The format of the strings generated by isoformat()
    # have a couple of problems:
    # 1) The strings generated by isotime are used in tokens and other public
    #    APIs that we can't change without a deprecation period. The strings
    #    generated by isoformat are not the same format, so we can't just
    #    change to it.
    # 2) The strings generated by isoformat do not include the microseconds if
    #    the value happens to be 0. This will likely show up as random failures
    #    as parsers may be written to always expect microseconds, and it will
    #    parse correctly most of the time.

    if not at:
        at = timeutils.utcnow()
    st = at.strftime(_ISO8601_TIME_FORMAT
                     if not subsecond
                     else _ISO8601_TIME_FORMAT_SUBSECOND)
    tz = at.tzinfo.tzname(None) if at.tzinfo else 'UTC'
    # Need to handle either iso8601 or python UTC format
    st += ('Z' if tz in ['UTC', 'UTC+00:00'] else tz)
    return st


def _get_root_helper():
    return 'sudo manila-rootwrap %s' % CONF.rootwrap_config


def execute(*cmd, **kwargs):
    """Convenience wrapper around oslo's execute() function."""
    if 'run_as_root' in kwargs and 'root_helper' not in kwargs:
        kwargs['root_helper'] = _get_root_helper()
    if hasattr('CONF', 'debug') and CONF.debug:
        kwargs['loglevel'] = logging.DEBUG
    return processutils.execute(*cmd, **kwargs)


class SSHPool(pools.Pool):
    """A simple eventlet pool to hold ssh connections."""

    def __init__(self, ip, port, conn_timeout, login, password=None,
                 privatekey=None, *args, **kwargs):
        self.ip = ip
        self.port = port
        self.login = login
        self.password = password
        self.conn_timeout = conn_timeout if conn_timeout else None
        self.path_to_private_key = privatekey
        super(SSHPool, self).__init__(*args, **kwargs)

    def create(self):  # pylint: disable=method-hidden
        ssh = paramiko.SSHClient()
        ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        look_for_keys = True
        if self.path_to_private_key:
            self.path_to_private_key = os.path.expanduser(
                self.path_to_private_key)
            look_for_keys = False
        elif self.password:
            look_for_keys = False
        try:
            LOG.debug("ssh.connect: ip: %s, port: %s, username: %s, "
                      "password: %s, key_filename: %s, look_for_keys: %s, "
                      "timeout: %s, banner_timeout: %s",
                      self.ip,
                      self.port,
                      self.login,
                      self.password,
                      self.path_to_private_key,
                      look_for_keys,
                      self.conn_timeout,
                      self.conn_timeout)
            ssh.connect(self.ip,
                        port=self.port,
                        username=self.login,
                        password=self.password,
                        key_filename=self.path_to_private_key,
                        look_for_keys=look_for_keys,
                        timeout=self.conn_timeout,
                        banner_timeout=self.conn_timeout)
            if self.conn_timeout:
                transport = ssh.get_transport()
                transport.set_keepalive(self.conn_timeout)
            return ssh
        except Exception as e:
            msg = _("Check whether private key or password are correctly "
                    "set. Error connecting via ssh: %s") % e
            LOG.error(msg)
            raise exception.SSHException(msg)

    def get(self):
        """Return an item from the pool, when one is available.

        This may cause the calling greenthread to block. Check if a
        connection is active before returning it. For dead connections
        create and return a new connection.
        """
        if self.free_items:
            conn = self.free_items.popleft()
            if conn:
                if conn.get_transport().is_active():
                    return conn
                else:
                    conn.close()
            return self.create()
        if self.current_size < self.max_size:
            created = self.create()
            self.current_size += 1
            return created
        return self.channel.get()

    def remove(self, ssh):
        """Close an ssh client and remove it from free_items."""
        ssh.close()
        if ssh in self.free_items:
            self.free_items.remove(ssh)
            if self.current_size > 0:
                self.current_size -= 1


def check_ssh_injection(cmd_list):
    ssh_injection_pattern = ['`', '$', '|', '||', ';', '&', '&&', '>', '>>',
                             '<']

    # Check whether injection attacks exist
    for arg in cmd_list:
        arg = arg.strip()

        # Check for matching quotes on the ends
        is_quoted = re.match('^(?P<quote>[\'"])(?P<quoted>.*)(?P=quote)$', arg)
        if is_quoted:
            # Check for unescaped quotes within the quoted argument
            quoted = is_quoted.group('quoted')
            if quoted:
                if (re.match('[\'"]', quoted) or
                        re.search('[^\\\\][\'"]', quoted)):
                    raise exception.SSHInjectionThreat(command=cmd_list)
        else:
            # We only allow spaces within quoted arguments, and that
            # is the only special character allowed within quotes
            if len(arg.split()) > 1:
                raise exception.SSHInjectionThreat(command=cmd_list)

        # Second, check whether danger character in command. So the shell
        # special operator must be a single argument.
        for c in ssh_injection_pattern:
            if c not in arg:
                continue

            result = arg.find(c)
            if not result == -1:
                if result == 0 or not arg[result - 1] == '\\':
                    raise exception.SSHInjectionThreat(command=cmd_list)


class LazyPluggable(object):
    """A pluggable backend loaded lazily based on some value."""

    def __init__(self, pivot, **backends):
        self.__backends = backends
        self.__pivot = pivot
        self.__backend = None

    def __get_backend(self):
        if not self.__backend:
            backend_name = CONF[self.__pivot]
            if backend_name not in self.__backends:
                raise exception.Error(_('Invalid backend: %s') % backend_name)

            backend = self.__backends[backend_name]
            if isinstance(backend, tuple):
                name = backend[0]
                fromlist = backend[1]
            else:
                name = backend
                fromlist = backend

            self.__backend = __import__(name, None, None, fromlist)
            LOG.debug('backend %s', self.__backend)
        return self.__backend

    def __getattr__(self, key):
        backend = self.__get_backend()
        return getattr(backend, key)


def monkey_patch():
    """Patch decorator.

    If the Flags.monkey_patch set as True,
    this function patches a decorator
    for all functions in specified modules.
    You can set decorators for each modules
    using CONF.monkey_patch_modules.
    The format is "Module path:Decorator function".
    Example: 'manila.api.ec2.cloud:' \
     manila.openstack.common.notifier.api.notify_decorator'

    Parameters of the decorator is as follows.
    (See manila.openstack.common.notifier.api.notify_decorator)

    name - name of the function
    function - object of the function
    """
    # If CONF.monkey_patch is not True, this function do nothing.
    if not CONF.monkey_patch:
        return
    # Get list of modules and decorators
    for module_and_decorator in CONF.monkey_patch_modules:
        module, decorator_name = module_and_decorator.split(':')
        # import decorator function
        decorator = importutils.import_class(decorator_name)
        __import__(module)
        # Retrieve module information using pyclbr
        module_data = pyclbr.readmodule_ex(module)
        for key in module_data.keys():
            # set the decorator for the class methods
            if isinstance(module_data[key], pyclbr.Class):
                clz = importutils.import_class("%s.%s" % (module, key))
                # NOTE(vponomaryov): we need to distinguish class methods types
                # for py2 and py3, because the concept of 'unbound methods' has
                # been removed from the python3.x
                if six.PY3:
                    member_type = inspect.isfunction
                else:
                    member_type = inspect.ismethod
                for method, func in inspect.getmembers(clz, member_type):
                    setattr(
                        clz, method,
                        decorator("%s.%s.%s" % (module, key, method), func))
            # set the decorator for the function
            if isinstance(module_data[key], pyclbr.Function):
                func = importutils.import_class("%s.%s" % (module, key))
                setattr(sys.modules[module], key,
                        decorator("%s.%s" % (module, key), func))


def file_open(*args, **kwargs):
    """Open file

    see built-in open() documentation for more details

    Note: The reason this is kept in a separate module is to easily
          be able to provide a stub module that doesn't alter system
          state at all (for unit tests)
    """
    return open(*args, **kwargs)


def service_is_up(service):
    """Check whether a service is up based on last heartbeat."""
    last_heartbeat = service['updated_at'] or service['created_at']
    # Timestamps in DB are UTC.
    tdelta = timeutils.utcnow() - last_heartbeat
    elapsed = tdelta.total_seconds()
    return abs(elapsed) <= CONF.service_down_time


def validate_service_host(context, host):
    service = db_api.service_get_by_host_and_topic(context, host,
                                                   'manila-share')
    if not service_is_up(service):
        raise exception.ServiceIsDown(service=service['host'])

    return service


@contextlib.contextmanager
def tempdir(**kwargs):
    tmpdir = tempfile.mkdtemp(**kwargs)
    try:
        yield tmpdir
    finally:
        try:
            shutil.rmtree(tmpdir)
        except OSError as e:
            LOG.debug('Could not remove tmpdir: %s', e)


def walk_class_hierarchy(clazz, encountered=None):
    """Walk class hierarchy, yielding most derived classes first."""
    if not encountered:
        encountered = []
    for subclass in clazz.__subclasses__():
        if subclass not in encountered:
            encountered.append(subclass)
            # drill down to leaves first
            for subsubclass in walk_class_hierarchy(subclass, encountered):
                yield subsubclass
            yield subclass


def cidr_to_network(cidr):
    """Convert cidr to network."""
    try:
        network = netaddr.IPNetwork(cidr)
        return network
    except netaddr.AddrFormatError:
        raise exception.InvalidInput(_("Invalid cidr supplied %s") % cidr)


def cidr_to_netmask(cidr):
    """Convert cidr to netmask."""
    return six.text_type(cidr_to_network(cidr).netmask)


def cidr_to_prefixlen(cidr):
    """Convert cidr to prefix length."""
    return cidr_to_network(cidr).prefixlen


def is_valid_ip_address(ip_address, ip_version):
    ip_version = ([int(ip_version)] if not isinstance(ip_version, list)
                  else ip_version)

    if not set(ip_version).issubset(set([4, 6])):
        raise exception.ManilaException(
            _("Provided improper IP version '%s'.") % ip_version)

    if 4 in ip_version:
        if netutils.is_valid_ipv4(ip_address):
            return True
    if 6 in ip_version:
        if netutils.is_valid_ipv6(ip_address):
            return True

    return False


def get_bool_param(param_string, params, default=False):
    param = params.get(param_string, default)
    if not strutils.is_valid_boolstr(param):
        msg = _("Value '%(param)s' for '%(param_string)s' is not "
                "a boolean.") % {'param': param, 'param_string': param_string}
        raise exception.InvalidParameterValue(err=msg)

    return strutils.bool_from_string(param, strict=True)


def is_all_tenants(search_opts):
    """Checks to see if the all_tenants flag is in search_opts

    :param dict search_opts: The search options for a request
    :returns: boolean indicating if all_tenants are being requested or not
    """
    all_tenants = search_opts.get('all_tenants')
    if all_tenants:
        try:
            all_tenants = strutils.bool_from_string(all_tenants, True)
        except ValueError as err:
            raise exception.InvalidInput(six.text_type(err))
    else:
        # The empty string is considered enabling all_tenants
        all_tenants = 'all_tenants' in search_opts
    return all_tenants


class IsAMatcher(object):
    def __init__(self, expected_value=None):
        self.expected_value = expected_value

    def __eq__(self, actual_value):
        return isinstance(actual_value, self.expected_value)


class ComparableMixin(object):
    def _compare(self, other, method):
        try:
            return method(self._cmpkey(), other._cmpkey())
        except (AttributeError, TypeError):
            # _cmpkey not implemented, or return different type,
            # so I can't compare with "other".
            return NotImplemented

    def __lt__(self, other):
        return self._compare(other, lambda s, o: s < o)

    def __le__(self, other):
        return self._compare(other, lambda s, o: s <= o)

    def __eq__(self, other):
        return self._compare(other, lambda s, o: s == o)

    def __ge__(self, other):
        return self._compare(other, lambda s, o: s >= o)

    def __gt__(self, other):
        return self._compare(other, lambda s, o: s > o)

    def __ne__(self, other):
        return self._compare(other, lambda s, o: s != o)


class retry_if_exit_code(tenacity.retry_if_exception):
    """Retry on ProcessExecutionError specific exit codes."""
    def __init__(self, codes):
        self.codes = (codes,) if isinstance(codes, int) else codes
        super(retry_if_exit_code, self).__init__(self._check_exit_code)

    def _check_exit_code(self, exc):
        return (exc and isinstance(exc, processutils.ProcessExecutionError) and
                exc.exit_code in self.codes)


def retry(retry_param=Exception,
          interval=1,
          retries=10,
          backoff_rate=2,
          backoff_sleep_max=None,
          wait_random=False,
          infinite=False,
          retry=tenacity.retry_if_exception_type):

    if retries < 1:
        raise ValueError('Retries must be greater than or '
                         'equal to 1 (received: %s). ' % retries)

    if wait_random:
        kwargs = {'multiplier': interval}
        if backoff_sleep_max is not None:
            kwargs.update({'max': backoff_sleep_max})
        wait = tenacity.wait_random_exponential(**kwargs)
    else:
        kwargs = {'multiplier': interval, 'min': 0, 'exp_base': backoff_rate}
        if backoff_sleep_max is not None:
            kwargs.update({'max': backoff_sleep_max})
        wait = tenacity.wait_exponential(**kwargs)

    if infinite:
        stop = tenacity.stop.stop_never
    else:
        stop = tenacity.stop_after_attempt(retries)

    def _decorator(f):

        @functools.wraps(f)
        def _wrapper(*args, **kwargs):
            r = tenacity.Retrying(
                sleep=tenacity.nap.sleep,
                before_sleep=tenacity.before_sleep_log(LOG, logging.DEBUG),
                after=tenacity.after_log(LOG, logging.DEBUG),
                stop=stop,
                reraise=True,
                retry=retry(retry_param),
                wait=wait)
            return r(f, *args, **kwargs)

        return _wrapper

    return _decorator


def get_bool_from_api_params(key, params, default=False, strict=True):
    """Parse bool value from request params.

    HTTPBadRequest will be directly raised either of the cases below:
    1. invalid bool string was found by key(with strict on).
    2. key not found while default value is invalid(with strict on).
    """
    param = params.get(key, default)
    try:
        param = strutils.bool_from_string(param,
                                          strict=strict,
                                          default=default)
    except ValueError:
        msg = _('Invalid value %(param)s for %(param_string)s. '
                'Expecting a boolean.') % {'param': param,
                                           'param_string': key}
        raise exc.HTTPBadRequest(explanation=msg)
    return param


def check_params_exist(keys, params):
    """Validates if keys exist in params.

    :param keys: List of keys to check
    :param params: Parameters received from REST API
    """
    if any(set(keys) - set(params)):
        msg = _("Must specify all mandatory parameters: %s") % keys
        raise exc.HTTPBadRequest(explanation=msg)


def check_params_are_boolean(keys, params, default=False):
    """Validates if keys in params are boolean.

    :param keys: List of keys to check
    :param params: Parameters received from REST API
    :param default: default value when it does not exist
    :return: a dictionary with keys and respective retrieved value
    """
    result = {}
    for key in keys:
        value = get_bool_from_api_params(key, params, default, strict=True)
        result[key] = value
    return result


def require_driver_initialized(func):
    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        # we can't do anything if the driver didn't init
        if not self.driver.initialized:
            driver_name = self.driver.__class__.__name__
            raise exception.DriverNotInitialized(driver=driver_name)
        return func(self, *args, **kwargs)
    return wrapper


def convert_str(text):
    """Convert to native string.

    Convert bytes and Unicode strings to native strings:

    * convert to bytes on Python 2:
      encode Unicode using encodeutils.safe_encode()
    * convert to Unicode on Python 3: decode bytes from UTF-8
    """
    if six.PY2:
        return encodeutils.safe_encode(text)
    else:
        if isinstance(text, bytes):
            return text.decode('utf-8')
        else:
            return text


def translate_string_size_to_float(string, multiplier='G'):
    """Translates human-readable storage size to float value.

    Supported values for 'multiplier' are following:
        K - kilo | 1
        M - mega | 1024
        G - giga | 1024 * 1024
        T - tera | 1024 * 1024 * 1024
        P = peta | 1024 * 1024 * 1024 * 1024

    returns:
        - float if correct input data provided
        - None if incorrect
    """
    if not isinstance(string, six.string_types):
        return None
    multipliers = ('K', 'M', 'G', 'T', 'P')
    mapping = {
        k: 1024.0 ** v
        for k, v in zip(multipliers, range(len(multipliers)))
    }
    if multiplier not in multipliers:
        raise exception.ManilaException(
            "'multiplier' arg should be one of following: "
            "'%(multipliers)s'. But it is '%(multiplier)s'." % {
                'multiplier': multiplier,
                'multipliers': "', '".join(multipliers),
            }
        )
    try:
        value = float(string.replace(",", ".")) / 1024.0
        value = value / mapping[multiplier]
        return value
    except (ValueError, TypeError):
        matched = re.match(
            r"^(\d*[.,]*\d*)([%s])$" % ''.join(multipliers), string)
        if matched:
            # The replace() is needed in case decimal separator is a comma
            value = float(matched.groups()[0].replace(",", "."))
            multiplier = mapping[matched.groups()[1]] / mapping[multiplier]
            return value * multiplier


def wait_for_access_update(context, db, share_instance,
                           migration_wait_access_rules_timeout):
    starttime = time.time()
    deadline = starttime + migration_wait_access_rules_timeout
    tries = 0

    while True:
        instance = db.share_instance_get(context, share_instance['id'])

        if instance['access_rules_status'] == constants.STATUS_ACTIVE:
            break

        tries += 1
        now = time.time()
        if (instance['access_rules_status'] ==
                constants.SHARE_INSTANCE_RULES_ERROR):
            msg = _("Failed to update access rules"
                    " on share instance %s") % share_instance['id']
            raise exception.ShareMigrationFailed(reason=msg)
        elif now > deadline:
            msg = _("Timeout trying to update access rules"
                    " on share instance %(share_id)s. Timeout "
                    "was %(timeout)s seconds.") % {
                'share_id': share_instance['id'],
                'timeout': migration_wait_access_rules_timeout}
            raise exception.ShareMigrationFailed(reason=msg)
        else:
            # 1.414 = square-root of 2
            time.sleep(1.414 ** tries)


class DoNothing(str):
    """Class that literrally does nothing.

    We inherit from str in case it's called with json.dumps.
    """

    def __call__(self, *args, **kwargs):
        return self

    def __getattr__(self, name):
        return self


DO_NOTHING = DoNothing()


def notifications_enabled(conf):
    """Check if oslo notifications are enabled."""
    notifications_driver = set(conf.oslo_messaging_notifications.driver)
    return notifications_driver and notifications_driver != {'noop'}


def if_notifications_enabled(function):
    """Calls decorated method only if notifications are enabled."""
    @functools.wraps(function)
    def wrapped(*args, **kwargs):
        if notifications_enabled(CONF):
            return function(*args, **kwargs)
        return DO_NOTHING
    return wrapped


def write_local_file(filename, contents, as_root=False):
    tmp_filename = "%s.tmp" % filename
    if as_root:
        execute('tee', tmp_filename, run_as_root=True, process_input=contents)
        execute('mv', '-f', tmp_filename, filename, run_as_root=True)
    else:
        with open(tmp_filename, 'w') as f:
            f.write(contents)
        os.rename(tmp_filename, filename)


def write_remote_file(ssh, filename, contents, as_root=False):
    tmp_filename = "%s.tmp" % filename
    if as_root:
        cmd = 'sudo tee "%s" > /dev/null' % tmp_filename
        cmd2 = 'sudo mv -f "%s" "%s"' % (tmp_filename, filename)
    else:
        cmd = 'cat > "%s"' % tmp_filename
        cmd2 = 'mv -f "%s" "%s"' % (tmp_filename, filename)
    stdin, __, __ = ssh.exec_command(cmd)
    stdin.write(contents)
    stdin.close()
    stdin.channel.shutdown_write()
    ssh.exec_command(cmd2)



# Copyright 2012 IBM Corp.
# Copyright (c) 2015 Mirantis inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import webob.exc

from manila.api.openstack import wsgi
from manila.api.views import services as services_views
from manila import db
from manila import utils


class ServiceMixin(object):
    """The Services API controller common logic.

    Mixin class that should be inherited by Services API controllers,
    which are used for different API URLs and microversions.
    """

    resource_name = "service"
    _view_builder_class = services_views.ViewBuilder

    @wsgi.Controller.authorize("index")
    def _index(self, req):
        """Return a list of all running services."""

        context = req.environ['manila.context']
        all_services = db.service_get_all(context)

        services = []
        for service in all_services:
            service = {
                'id': service['id'],
                'binary': service['binary'],
                'host': service['host'],
                'zone': service['availability_zone']['name'],
                'status': 'disabled' if service['disabled'] else 'enabled',
                'state': 'up' if utils.service_is_up(service) else 'down',
                'updated_at': service['updated_at'],
            }
            services.append(service)

        search_opts = [
            'host',
            'binary',
            'zone',
            'state',
            'status',
        ]
        for search_opt in search_opts:
            if search_opt in req.GET:
                value = req.GET[search_opt]
                services = [s for s in services if s[search_opt] == value]
            if len(services) == 0:
                break

        return self._view_builder.detail_list(services)

    @wsgi.Controller.authorize("update")
    def _update(self, req, id, body):
        """Enable/Disable scheduling for a service."""
        context = req.environ['manila.context']

        if id == "enable":
            data = {'disabled': False}
        elif id == "disable":
            data = {'disabled': True}
        else:
            raise webob.exc.HTTPNotFound("Unknown action '%s'" % id)

        try:
            data['host'] = body['host']
            data['binary'] = body['binary']
        except (TypeError, KeyError):
            raise webob.exc.HTTPBadRequest()

        svc = db.service_get_by_args(context, data['host'], data['binary'])
        db.service_update(
            context, svc['id'], {'disabled': data['disabled']})

        return self._view_builder.summary(data)


class ServiceControllerLegacy(ServiceMixin, wsgi.Controller):
    """Deprecated Services API controller.

    Used by legacy API v1 and v2 microversions from 2.0 to 2.6.
    Registered under deprecated API URL 'os-services'.
    """

    @wsgi.Controller.api_version('1.0', '2.6')
    def index(self, req):
        return self._index(req)

    @wsgi.Controller.api_version('1.0', '2.6')
    def update(self, req, id, body):
        return self._update(req, id, body)


class ServiceController(ServiceMixin, wsgi.Controller):
    """Services API controller.

    Used only by API v2 starting from microversion 2.7.
    Registered under API URL 'services'.
    """

    @wsgi.Controller.api_version('2.7')
    def index(self, req):
        return self._index(req)

    @wsgi.Controller.api_version('2.7')
    def update(self, req, id, body):
        return self._update(req, id, body)


def create_resource_legacy():
    return wsgi.Resource(ServiceControllerLegacy())


def create_resource():
    return wsgi.Resource(ServiceController())



# Copyright (c) 2016 Clinton Knight.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Add 'create_share_from_snapshot_support' extra spec to share types

Revision ID: 3e7d62517afa
Revises: 48a7beae3117
Create Date: 2016-08-16 10:48:11.497499

"""

# revision identifiers, used by Alembic.
revision = '3e7d62517afa'
down_revision = '48a7beae3117'

from alembic import op
from oslo_utils import timeutils
import sqlalchemy as sa
from sqlalchemy.sql import table

from manila.common import constants


def upgrade():
    """Performs DB upgrade to add create_share_from_snapshot_support.

    Prior to this migration, the 'snapshot_support' extra spec meant two
    different things: a snapshot may be created, and a new share may be created
    from a snapshot. With the planned addition of new snapshot semantics
    (revert to snapshot, mountable snapshots), it is likely a driver may be
    able to support one or both of the new semantics but *not* be able to
    create a share from a snapshot. So this migration separates the existing
    snapshot_support extra spec and share attribute into two values to enable
    logical separability of the features.

    Add 'create_share_from_snapshot_support' extra spec to all share types and
    attribute 'create_share_from_snapshot_support' to Share model.
    """
    session = sa.orm.Session(bind=op.get_bind().connect())

    extra_specs_table = table(
        'share_type_extra_specs',
        sa.Column('created_at', sa.DateTime),
        sa.Column('deleted', sa.Integer),
        sa.Column('share_type_id', sa.String(length=36)),
        sa.Column('spec_key', sa.String(length=255)),
        sa.Column('spec_value', sa.String(length=255)))

    share_type_table = table(
        'share_types',
        sa.Column('deleted', sa.Integer),
        sa.Column('id', sa.Integer))

    # Get list of share type IDs that don't already have the new required
    # create_share_from_snapshot_support extra spec defined.
    existing_extra_specs = session.query(
        extra_specs_table).filter(
        extra_specs_table.c.spec_key ==
        constants.ExtraSpecs.CREATE_SHARE_FROM_SNAPSHOT_SUPPORT).filter(
        extra_specs_table.c.deleted == 0).all()
    excluded_st_ids = [es.share_type_id for es in existing_extra_specs]

    # Get share types for the IDs we got in the previous query
    share_types = session.query(share_type_table).filter(
        share_type_table.c.deleted.in_(('0', 'False', ))).filter(
        share_type_table.c.id.notin_(excluded_st_ids)).all()

    extra_specs = []
    now = timeutils.utcnow()
    for share_type in share_types:

        # Get the value of snapshot_support for each extant share type
        snapshot_support_extra_spec = session.query(
            extra_specs_table).filter(
            extra_specs_table.c.spec_key ==
            constants.ExtraSpecs.SNAPSHOT_SUPPORT).filter(
            extra_specs_table.c.share_type_id == share_type.id).first()

        spec_value = (snapshot_support_extra_spec.spec_value if
                      snapshot_support_extra_spec else 'False')

        # Copy the snapshot_support value to create_share_from_snapshot_support
        extra_specs.append({
            'spec_key':
            constants.ExtraSpecs.CREATE_SHARE_FROM_SNAPSHOT_SUPPORT,
            'spec_value': spec_value,
            'deleted': 0,
            'created_at': now,
            'share_type_id': share_type.id,
        })
    if extra_specs:
        op.bulk_insert(extra_specs_table, extra_specs)

    # Add create_share_from_snapshot_support attribute to shares table
    op.add_column('shares',
                  sa.Column('create_share_from_snapshot_support',
                            sa.Boolean, default=True))

    # Copy snapshot_support to create_share_from_snapshot_support on each share
    shares_table = sa.Table(
        'shares',
        sa.MetaData(),
        sa.Column('id', sa.String(length=36)),
        sa.Column('deleted', sa.String(length=36)),
        sa.Column('snapshot_support', sa.Boolean),
        sa.Column('create_share_from_snapshot_support', sa.Boolean),
    )
    # pylint: disable=no-value-for-parameter
    update = shares_table.update().where(
        shares_table.c.deleted == 'False').values(
        create_share_from_snapshot_support=shares_table.c.snapshot_support)
    session.execute(update)
    session.commit()

    session.close_all()


def downgrade():
    """Performs DB downgrade removing create_share_from_snapshot_support.

    Remove 'create_share_from_snapshot_support' extra spec from all share types
    and attribute 'create_share_from_snapshot_support' from Share model.
    """
    connection = op.get_bind().connect()
    deleted_at = timeutils.utcnow()
    extra_specs = sa.Table(
        'share_type_extra_specs',
        sa.MetaData(),
        autoload=True,
        autoload_with=connection)

    # pylint: disable=no-value-for-parameter
    update = extra_specs.update().where(
        extra_specs.c.spec_key ==
        constants.ExtraSpecs.CREATE_SHARE_FROM_SNAPSHOT_SUPPORT).where(
        extra_specs.c.deleted == 0).values(deleted=extra_specs.c.id,
                                           deleted_at=deleted_at)
    connection.execute(update)

    op.drop_column('shares', 'create_share_from_snapshot_support')



# Copyright 2018 Huawei Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""The share accesses api."""

import ast

import webob

from manila.api.openstack import wsgi
from manila.api.views import share_accesses as share_access_views
from manila import exception
from manila.i18n import _
from manila import share


class ShareAccessesController(wsgi.Controller, wsgi.AdminActionsMixin):
    """The Share accesses API V2 controller for the OpenStack API."""

    resource_name = 'share_access_rule'
    _view_builder_class = share_access_views.ViewBuilder

    def __init__(self):
        super(ShareAccessesController, self).__init__()
        self.share_api = share.API()

    @wsgi.Controller.api_version('2.45')
    @wsgi.Controller.authorize('get')
    def show(self, req, id):
        """Return data about the given share access rule."""
        context = req.environ['manila.context']
        share_access = self._get_share_access(context, id)
        return self._view_builder.view(req, share_access)

    def _get_share_access(self, context, share_access_id):
        try:
            return self.share_api.access_get(context, share_access_id)
        except exception.NotFound:
            msg = _("Share access rule %s not found.") % share_access_id
            raise webob.exc.HTTPNotFound(explanation=msg)

    @wsgi.Controller.api_version('2.45')
    @wsgi.Controller.authorize
    def index(self, req):
        """Returns the list of access rules for a given share."""
        context = req.environ['manila.context']
        search_opts = {}
        search_opts.update(req.GET)
        if 'share_id' not in search_opts:
            msg = _("The field 'share_id' has to be specified.")
            raise webob.exc.HTTPBadRequest(explanation=msg)
        share_id = search_opts.pop('share_id', None)

        if 'metadata' in search_opts:
            search_opts['metadata'] = ast.literal_eval(
                search_opts['metadata'])
        try:
            share = self.share_api.get(context, share_id)
        except exception.NotFound:
            msg = _("Share %s not found.") % share_id
            raise webob.exc.HTTPBadRequest(explanation=msg)
        access_rules = self.share_api.access_get_all(
            context, share, search_opts)

        return self._view_builder.list_view(req, access_rules)


def create_resource():
    return wsgi.Resource(ShareAccessesController())



# Copyright (c) 2016 Hitachi Data Systems, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import ddt
import json
import requests
import time
from unittest import mock

from manila import exception
from manila.share.drivers.hitachi.hsp import rest
from manila import test
from manila.tests.share.drivers.hitachi.hsp import fakes


class FakeRequests(object):
    status_code = 0
    headers = {}
    content = ""

    def __init__(self, status_code, content='null'):
        self.status_code = status_code
        self.headers = {'location': 'fake_location'}
        self.content = content

    def json(self):
        return {'messages': [{'message': 'fake_msg'}]}


@ddt.ddt
class HitachiHSPRestTestCase(test.TestCase):
    def setUp(self):
        super(HitachiHSPRestTestCase, self).setUp()
        self.hitachi_hsp_host = '172.24.47.190'
        self.hitachi_hsp_username = 'hds_hnas_user'
        self.hitachi_hsp_password = 'hds_hnas_password'

        self._driver = rest.HSPRestBackend(self.hitachi_hsp_host,
                                           self.hitachi_hsp_username,
                                           self.hitachi_hsp_password)

    @ddt.data(202, 500)
    def test__send_post(self, code):
        self.mock_object(requests, "post", mock.Mock(
            return_value=FakeRequests(code)))

        if code == 202:
            self.mock_object(rest.HSPRestBackend, "_wait_job_status",
                             mock.Mock())
            self._driver._send_post('fake_url')

            rest.HSPRestBackend._wait_job_status.assert_called_once_with(
                'fake_location', 'COMPLETE')
        else:
            self.assertRaises(exception.HSPBackendException,
                              self._driver._send_post, 'fake_url')

    @ddt.data({'code': 200, 'content': 'null'},
              {'code': 200, 'content': 'fake_content'},
              {'code': 500, 'content': 'null'})
    @ddt.unpack
    def test__send_get(self, code, content):
        self.mock_object(requests, "get", mock.Mock(
            return_value=FakeRequests(code, content)))

        if code == 200:
            result = self._driver._send_get('fake_url')
            if content == 'null':
                self.assertIsNone(result)
            else:
                self.assertEqual(FakeRequests(code, content).json(), result)
        else:
            self.assertRaises(exception.HSPBackendException,
                              self._driver._send_get, 'fake_url')

    @ddt.data(202, 500)
    def test__send_delete(self, code):
        self.mock_object(requests, "delete", mock.Mock(
            return_value=FakeRequests(code)))

        if code == 202:
            self.mock_object(rest.HSPRestBackend, "_wait_job_status",
                             mock.Mock())
            self._driver._send_delete('fake_url')

            rest.HSPRestBackend._wait_job_status.assert_called_once_with(
                'fake_location', 'COMPLETE')
        else:
            self.assertRaises(exception.HSPBackendException,
                              self._driver._send_delete, 'fake_url')

    def test_add_file_system(self):
        url = "https://172.24.47.190/hspapi/file-systems/"

        payload = {
            'quota': fakes.file_system['properties']['quota'],
            'auto-access': False,
            'enabled': True,
            'description': '',
            'record-access-time': True,
            'tags': '',
            'space-hwm': 90,
            'space-lwm': 70,
            'name': fakes.file_system['properties']['name'],
        }

        self.mock_object(rest.HSPRestBackend, "_send_post", mock.Mock())
        self._driver.add_file_system(fakes.file_system['properties']['name'],
                                     fakes.file_system['properties']['quota'])

        rest.HSPRestBackend._send_post.assert_called_once_with(
            url, payload=json.dumps(payload))

    def test_get_file_system(self):
        url = ("https://172.24.47.190/hspapi/file-systems/list?name=%s" %
               fakes.file_system['properties']['name'])

        self.mock_object(rest.HSPRestBackend, "_send_get", mock.Mock(
            return_value={'list': [fakes.file_system]}))

        result = self._driver.get_file_system(
            fakes.file_system['properties']['name'])

        self.assertEqual(fakes.file_system, result)

        rest.HSPRestBackend._send_get.assert_called_once_with(url)

    def test_get_file_system_exception(self):
        url = ("https://172.24.47.190/hspapi/file-systems/list?name=%s" %
               fakes.file_system['properties']['name'])

        self.mock_object(rest.HSPRestBackend, "_send_get",
                         mock.Mock(return_value=None))

        self.assertRaises(exception.HSPItemNotFoundException,
                          self._driver.get_file_system,
                          fakes.file_system['properties']['name'])

        rest.HSPRestBackend._send_get.assert_called_once_with(url)

    def test_delete_file_system(self):
        url = ("https://172.24.47.190/hspapi/file-systems/%s" %
               fakes.file_system['id'])

        self.mock_object(rest.HSPRestBackend, "_send_delete", mock.Mock())
        self._driver.delete_file_system(fakes.file_system['id'])

        rest.HSPRestBackend._send_delete.assert_called_once_with(url)

    def test_resize_file_system(self):
        url = ("https://172.24.47.190/hspapi/file-systems/%s" %
               fakes.file_system['id'])
        new_size = 53687091200
        payload = {'quota': new_size}

        self.mock_object(rest.HSPRestBackend, "_send_post", mock.Mock())
        self._driver.resize_file_system(fakes.file_system['id'], new_size)

        rest.HSPRestBackend._send_post.assert_called_once_with(
            url, payload=json.dumps(payload))

    def test_rename_file_system(self):
        url = ("https://172.24.47.190/hspapi/file-systems/%s" %
               fakes.file_system['id'])
        new_name = "fs_rename"
        payload = {'name': new_name}

        self.mock_object(rest.HSPRestBackend, "_send_post", mock.Mock())

        self._driver.rename_file_system(fakes.file_system['id'], new_name)

        rest.HSPRestBackend._send_post.assert_called_once_with(
            url, payload=json.dumps(payload))

    def test_add_share(self):
        url = "https://172.24.47.190/hspapi/shares/"
        payload = {
            'description': '',
            'type': 'NFS',
            'enabled': True,
            'tags': '',
            'name': fakes.share['name'],
            'file-system-id': fakes.share['properties']['file-system-id'],
        }

        self.mock_object(rest.HSPRestBackend, "_send_post", mock.Mock())

        self._driver.add_share(fakes.share['name'],
                               fakes.share['properties']['file-system-id'])

        rest.HSPRestBackend._send_post.assert_called_once_with(
            url, payload=json.dumps(payload))

    @ddt.data({'fs_id': None,
               'name': fakes.share['name'],
               'url': 'https://172.24.47.190/hspapi/shares/list?'
                      'name=aa4a7710-f326-41fb-ad18-b4ad587fc87a'},
              {'fs_id': fakes.share['properties']['file-system-id'],
               'name': None,
               'url': 'https://172.24.47.190/hspapi/shares/list?'
                      'file-system-id=33689245-1806-45d0-8507-0700b5f89750'})
    @ddt.unpack
    def test_get_share(self, fs_id, name, url):
        self.mock_object(rest.HSPRestBackend, "_send_get",
                         mock.Mock(return_value={'list': [fakes.share]}))

        result = self._driver.get_share(fs_id, name)

        self.assertEqual(fakes.share, result)

        rest.HSPRestBackend._send_get.assert_called_once_with(url)

    def test_get_share_exception(self):
        url = ("https://172.24.47.190/hspapi/shares/list?"
               "name=aa4a7710-f326-41fb-ad18-b4ad587fc87a")

        self.mock_object(rest.HSPRestBackend, "_send_get", mock.Mock(
            return_value=None))

        self.assertRaises(exception.HSPItemNotFoundException,
                          self._driver.get_share, None, fakes.share['name'])

        rest.HSPRestBackend._send_get.assert_called_once_with(url)

    def test_delete_share(self):
        url = "https://172.24.47.190/hspapi/shares/%s" % fakes.share['id']

        self.mock_object(rest.HSPRestBackend, "_send_delete")

        self._driver.delete_share(fakes.share['id'])

        rest.HSPRestBackend._send_delete.assert_called_once_with(url)

    def test_add_access_rule(self):
        url = "https://172.24.47.190/hspapi/shares/%s/" % fakes.share['id']
        payload = {
            "action": "add-access-rule",
            "name": fakes.share['id'] + fakes.access_rule['access_to'],
            "host-specification": fakes.access_rule['access_to'],
            "read-write": fakes.access_rule['access_level'],
        }

        self.mock_object(rest.HSPRestBackend, "_send_post", mock.Mock())

        self._driver.add_access_rule(fakes.share['id'],
                                     fakes.access_rule['access_to'],
                                     fakes.access_rule['access_level'])

        rest.HSPRestBackend._send_post.assert_called_once_with(
            url, payload=json.dumps(payload))

    def test_delete_access_rule(self):
        url = "https://172.24.47.190/hspapi/shares/%s/" % fakes.share['id']
        payload = {
            "action": "delete-access-rule",
            "name": fakes.hsp_rules[0]['name'],
        }
        self.mock_object(rest.HSPRestBackend, "_send_post", mock.Mock())

        self._driver.delete_access_rule(fakes.share['id'],
                                        fakes.hsp_rules[0]['name'])

        rest.HSPRestBackend._send_post.assert_called_once_with(
            url, payload=json.dumps(payload))

    @ddt.data({'value': {'list': fakes.hsp_rules}, 'res': fakes.hsp_rules},
              {'value': None, 'res': []})
    @ddt.unpack
    def test_get_access_rules(self, value, res):
        url = ("https://172.24.47.190/hspapi/shares/%s/access-rules" %
               fakes.share['id'])

        self.mock_object(rest.HSPRestBackend, "_send_get", mock.Mock(
            return_value=value))

        result = self._driver.get_access_rules(fakes.share['id'])

        self.assertEqual(res, result)

        rest.HSPRestBackend._send_get.assert_called_once_with(url)

    @ddt.data({'list': [fakes.hsp_cluster]}, None)
    def test_get_clusters(self, value):
        url = "https://172.24.47.190/hspapi/clusters/list"

        self.mock_object(rest.HSPRestBackend, "_send_get", mock.Mock(
            return_value=value))

        if value:
            result = self._driver.get_cluster()

            self.assertEqual(fakes.hsp_cluster, result)
        else:
            self.assertRaises(exception.HSPBackendException,
                              self._driver.get_cluster)

        rest.HSPRestBackend._send_get.assert_called_once_with(url)

    @ddt.data('COMPLETE', 'ERROR', 'RUNNING')
    def test__wait_job_status(self, stat):
        url = "fake_job_url"
        json = {
            'id': 'fake_id',
            'properties': {
                'completion-details': 'Duplicate NFS access rule exists',
                'completion-status': stat,
            },
            'messages': [{
                'id': 'fake_id',
                'message': 'fake_msg',
            }]
        }

        self.mock_object(rest.HSPRestBackend, "_send_get", mock.Mock(
            return_value=json))
        self.mock_object(time, "sleep")

        if stat == 'COMPLETE':
            self._driver._wait_job_status(url, 'COMPLETE')

            rest.HSPRestBackend._send_get.assert_called_once_with(url)
        elif stat == 'ERROR':
            self.assertRaises(exception.HSPBackendException,
                              self._driver._wait_job_status, url, 'COMPLETE')

            rest.HSPRestBackend._send_get.assert_called_once_with(url)
        else:
            self.assertRaises(exception.HSPTimeoutException,
                              self._driver._wait_job_status, url, 'COMPLETE')

            rest.HSPRestBackend._send_get.assert_has_calls([
                mock.call(url), mock.call(url), mock.call(url), mock.call(url),
                mock.call(url),
            ])



# Copyright (c) 2016 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from manila.api import common
from manila.common import constants
from manila.share import api as share_api


class ViewBuilder(common.ViewBuilder):
    """Model a share access API response as a python dictionary."""

    _collection_name = 'share_accesses'
    _detail_version_modifiers = [
        "add_access_key",
        "translate_transitional_statuses",
        "add_created_at_and_updated_at",
        "add_access_rule_metadata_field",
    ]

    def list_view(self, request, accesses):
        """View of a list of share accesses."""
        return {'access_list': [self.summary_view(request, access)['access']
                                for access in accesses]}

    def summary_view(self, request, access):
        """Summarized view of a single share access."""
        access_dict = {
            'id': access.get('id'),
            'access_level': access.get('access_level'),
            'access_to': access.get('access_to'),
            'access_type': access.get('access_type'),
            'state': access.get('state'),
        }
        self.update_versioned_resource_dict(
            request, access_dict, access)
        return {'access': access_dict}

    def view(self, request, access):
        """Generic view of a single share access."""
        access_dict = {
            'id': access.get('id'),
            'share_id': access.get('share_id'),
            'access_level': access.get('access_level'),
            'access_to': access.get('access_to'),
            'access_type': access.get('access_type'),
            'state': access.get('state'),
        }
        self.update_versioned_resource_dict(
            request, access_dict, access)
        return {'access': access_dict}

    def view_metadata(self, request, metadata):
        """View of a share access rule metadata."""
        return {'metadata': metadata}

    @common.ViewBuilder.versioned_method("2.21")
    def add_access_key(self, context, access_dict, access):
        access_dict['access_key'] = access.get('access_key')

    @common.ViewBuilder.versioned_method("2.33")
    def add_created_at_and_updated_at(self, context, access_dict, access):
        access_dict['created_at'] = access.get('created_at')
        access_dict['updated_at'] = access.get('updated_at')

    @common.ViewBuilder.versioned_method("2.45")
    def add_access_rule_metadata_field(self, context, access_dict, access):
        metadata = access.get('share_access_rules_metadata') or {}
        metadata = {item['key']: item['value'] for item in metadata}
        access_dict['metadata'] = metadata

    @common.ViewBuilder.versioned_method("1.0", "2.27")
    def translate_transitional_statuses(self, context, access_dict, access):
        """In 2.28, the per access rule status was (re)introduced."""
        api = share_api.API()
        share = api.get(context, access['share_id'])

        if (share['access_rules_status'] ==
                constants.SHARE_INSTANCE_RULES_SYNCING):
            access_dict['state'] = constants.STATUS_NEW
        else:
            access_dict['state'] = share['access_rules_status']



# Copyright 2011 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import ddt
from oslo_config import cfg
from oslo_serialization import jsonutils
import webob

from manila.api.v1 import share_metadata
from manila.api.v1 import shares
from manila import context
from manila import db
from manila.share import api
from manila import test
from manila.tests.api import fakes

CONF = cfg.CONF


@ddt.ddt
class ShareMetaDataTest(test.TestCase):

    def setUp(self):
        super(ShareMetaDataTest, self).setUp()
        self.share_api = api.API()
        self.share_controller = shares.ShareController()
        self.controller = share_metadata.ShareMetadataController()
        self.ctxt = context.RequestContext('admin', 'fake', True)
        self.origin_metadata = {
            "key1": "value1",
            "key2": "value2",
            "key3": "value3",
        }
        self.share = db.share_create(self.ctxt, {})
        self.share_id = self.share['id']
        self.url = '/shares/%s/metadata' % self.share_id
        db.share_metadata_update(
            self.ctxt, self.share_id, self.origin_metadata, delete=False)

    def test_index(self):
        req = fakes.HTTPRequest.blank(self.url)
        res_dict = self.controller.index(req, self.share_id)

        expected = {
            'metadata': {
                'key1': 'value1',
                'key2': 'value2',
                'key3': 'value3',
            },
        }
        self.assertEqual(expected, res_dict)

    def test_index_nonexistent_share(self):
        req = fakes.HTTPRequest.blank(self.url)
        self.assertRaises(webob.exc.HTTPNotFound,
                          self.controller.index, req, self.url)

    def test_index_no_data(self):
        db.share_metadata_update(
            self.ctxt, self.share_id, {}, delete=True)
        req = fakes.HTTPRequest.blank(self.url)
        res_dict = self.controller.index(req, self.share_id)
        expected = {'metadata': {}}
        self.assertEqual(expected, res_dict)

    def test_show(self):
        req = fakes.HTTPRequest.blank(self.url + '/key2')

        res_dict = self.controller.show(req, self.share_id, 'key2')

        expected = {'meta': {'key2': 'value2'}}
        self.assertEqual(expected, res_dict)

    def test_show_nonexistent_share(self):
        req = fakes.HTTPRequest.blank(self.url + '/key2')
        self.assertRaises(
            webob.exc.HTTPNotFound,
            self.controller.show,
            req, "nonexistent_share", 'key2')

    def test_show_meta_not_found(self):
        req = fakes.HTTPRequest.blank(self.url + '/key6')
        self.assertRaises(webob.exc.HTTPNotFound,
                          self.controller.show, req, self.share_id, 'key6')

    def test_delete(self):
        req = fakes.HTTPRequest.blank(self.url + '/key2')
        req.method = 'DELETE'
        res = self.controller.delete(req, self.share_id, 'key2')

        self.assertEqual(200, res.status_int)

    def test_delete_nonexistent_share(self):
        req = fakes.HTTPRequest.blank(self.url + '/key1')
        req.method = 'DELETE'
        self.assertRaises(
            webob.exc.HTTPNotFound,
            self.controller.delete,
            req, "nonexistent_share", 'key1')

    def test_delete_meta_not_found(self):
        req = fakes.HTTPRequest.blank(self.url + '/key6')
        req.method = 'DELETE'
        self.assertRaises(webob.exc.HTTPNotFound,
                          self.controller.delete, req, self.share_id, 'key6')

    def test_create(self):
        req = fakes.HTTPRequest.blank('/v1/share_metadata')
        req.method = 'POST'
        req.content_type = "application/json"
        body = {"metadata": {"key9": "value9"}}
        req.body = jsonutils.dumps(body).encode("utf-8")
        res_dict = self.controller.create(req, self.share_id, body)
        expected = self.origin_metadata
        expected.update(body['metadata'])
        self.assertEqual({'metadata': expected}, res_dict)

    def test_create_empty_body(self):
        req = fakes.HTTPRequest.blank(self.url)
        req.method = 'POST'
        req.headers["content-type"] = "application/json"

        self.assertRaises(webob.exc.HTTPBadRequest,
                          self.controller.create, req, self.share_id, None)

    def test_create_item_empty_key(self):
        req = fakes.HTTPRequest.blank(self.url + '/key1')
        req.method = 'PUT'
        body = {"meta": {"": "value1"}}
        req.body = jsonutils.dumps(body).encode("utf-8")
        req.headers["content-type"] = "application/json"

        self.assertRaises(webob.exc.HTTPBadRequest,
                          self.controller.create, req, self.share_id, body)

    def test_create_item_key_too_long(self):
        req = fakes.HTTPRequest.blank(self.url + '/key1')
        req.method = 'PUT'
        body = {"meta": {("a" * 260): "value1"}}
        req.body = jsonutils.dumps(body).encode("utf-8")
        req.headers["content-type"] = "application/json"

        self.assertRaises(webob.exc.HTTPBadRequest,
                          self.controller.create,
                          req, self.share_id, body)

    def test_create_nonexistent_share(self):
        req = fakes.HTTPRequest.blank('/v1/share_metadata')
        req.method = 'POST'
        req.content_type = "application/json"
        body = {"metadata": {"key9": "value9"}}
        req.body = jsonutils.dumps(body).encode("utf-8")
        self.assertRaises(
            webob.exc.HTTPNotFound,
            self.controller.create,
            req, "nonexistent_share", body)

    def test_update_all(self):
        req = fakes.HTTPRequest.blank(self.url)
        req.method = 'PUT'
        req.content_type = "application/json"
        expected = {
            'metadata': {
                'key10': 'value10',
                'key99': 'value99',
            },
        }
        req.body = jsonutils.dumps(expected).encode("utf-8")
        res_dict = self.controller.update_all(req, self.share_id, expected)

        self.assertEqual(expected, res_dict)

    def test_update_all_empty_container(self):
        req = fakes.HTTPRequest.blank(self.url)
        req.method = 'PUT'
        req.content_type = "application/json"
        expected = {'metadata': {}}
        req.body = jsonutils.dumps(expected).encode("utf-8")
        res_dict = self.controller.update_all(req, self.share_id, expected)

        self.assertEqual(expected, res_dict)

    def test_update_all_malformed_container(self):
        req = fakes.HTTPRequest.blank(self.url)
        req.method = 'PUT'
        req.content_type = "application/json"
        expected = {'meta': {}}
        req.body = jsonutils.dumps(expected).encode("utf-8")

        self.assertRaises(webob.exc.HTTPBadRequest,
                          self.controller.update_all, req, self.share_id,
                          expected)

    @ddt.data(['asdf'],
              {'key': None},
              {None: 'value'},
              {None: None})
    def test_update_all_malformed_data(self, metadata):
        req = fakes.HTTPRequest.blank(self.url)
        req.method = 'PUT'
        req.content_type = "application/json"
        expected = {'metadata': metadata}
        req.body = jsonutils.dumps(expected).encode("utf-8")

        self.assertRaises(webob.exc.HTTPBadRequest,
                          self.controller.update_all, req, self.share_id,
                          expected)

    def test_update_all_nonexistent_share(self):
        req = fakes.HTTPRequest.blank(self.url)
        req.method = 'PUT'
        req.content_type = "application/json"
        body = {'metadata': {'key10': 'value10'}}
        req.body = jsonutils.dumps(body).encode("utf-8")

        self.assertRaises(webob.exc.HTTPNotFound,
                          self.controller.update_all, req, '100', body)

    def test_update_item(self):
        req = fakes.HTTPRequest.blank(self.url + '/key1')
        req.method = 'PUT'
        body = {"meta": {"key1": "value1"}}
        req.body = jsonutils.dumps(body).encode("utf-8")
        req.headers["content-type"] = "application/json"
        res_dict = self.controller.update(req, self.share_id, 'key1', body)
        expected = {'meta': {'key1': 'value1'}}
        self.assertEqual(expected, res_dict)

    def test_update_item_nonexistent_share(self):
        req = fakes.HTTPRequest.blank('/v1.1/fake/shares/asdf/metadata/key1')
        req.method = 'PUT'
        body = {"meta": {"key1": "value1"}}
        req.body = jsonutils.dumps(body).encode("utf-8")
        req.headers["content-type"] = "application/json"

        self.assertRaises(
            webob.exc.HTTPNotFound,
            self.controller.update,
            req, "nonexistent_share", 'key1', body)

    def test_update_item_empty_body(self):
        req = fakes.HTTPRequest.blank(self.url + '/key1')
        req.method = 'PUT'
        req.headers["content-type"] = "application/json"

        self.assertRaises(webob.exc.HTTPBadRequest,
                          self.controller.update, req, self.share_id, 'key1',
                          None)

    def test_update_item_empty_key(self):
        req = fakes.HTTPRequest.blank(self.url + '/key1')
        req.method = 'PUT'
        body = {"meta": {"": "value1"}}
        req.body = jsonutils.dumps(body).encode("utf-8")
        req.headers["content-type"] = "application/json"

        self.assertRaises(webob.exc.HTTPBadRequest,
                          self.controller.update, req, self.share_id, '', body)

    def test_update_item_key_too_long(self):
        req = fakes.HTTPRequest.blank(self.url + '/key1')
        req.method = 'PUT'
        body = {"meta": {("a" * 260): "value1"}}
        req.body = jsonutils.dumps(body).encode("utf-8")
        req.headers["content-type"] = "application/json"

        self.assertRaises(webob.exc.HTTPBadRequest,
                          self.controller.update,
                          req, self.share_id, ("a" * 260), body)

    def test_update_item_value_too_long(self):
        req = fakes.HTTPRequest.blank(self.url + '/key1')
        req.method = 'PUT'
        body = {"meta": {"key1": ("a" * 1025)}}
        req.body = jsonutils.dumps(body).encode("utf-8")
        req.headers["content-type"] = "application/json"

        self.assertRaises(webob.exc.HTTPBadRequest,
                          self.controller.update,
                          req, self.share_id, "key1", body)

    def test_update_item_too_many_keys(self):
        req = fakes.HTTPRequest.blank(self.url + '/key1')
        req.method = 'PUT'
        body = {"meta": {"key1": "value1", "key2": "value2"}}
        req.body = jsonutils.dumps(body).encode("utf-8")
        req.headers["content-type"] = "application/json"

        self.assertRaises(webob.exc.HTTPBadRequest,
                          self.controller.update, req, self.share_id, 'key1',
                          body)

    def test_update_item_body_uri_mismatch(self):
        req = fakes.HTTPRequest.blank(self.url + '/bad')
        req.method = 'PUT'
        body = {"meta": {"key1": "value1"}}
        req.body = jsonutils.dumps(body).encode("utf-8")
        req.headers["content-type"] = "application/json"

        self.assertRaises(webob.exc.HTTPBadRequest,
                          self.controller.update, req, self.share_id, 'bad',
                          body)

    def test_invalid_metadata_items_on_create(self):
        req = fakes.HTTPRequest.blank(self.url)
        req.method = 'POST'
        req.headers["content-type"] = "application/json"

        # test for long key
        data = {"metadata": {"a" * 260: "value1"}}
        req.body = jsonutils.dumps(data).encode("utf-8")
        self.assertRaises(webob.exc.HTTPBadRequest,
                          self.controller.create, req, self.share_id, data)

        # test for long value
        data = {"metadata": {"key": "v" * 1025}}
        req.body = jsonutils.dumps(data).encode("utf-8")
        self.assertRaises(webob.exc.HTTPBadRequest,
                          self.controller.create, req, self.share_id, data)

        # test for empty key.
        data = {"metadata": {"": "value1"}}
        req.body = jsonutils.dumps(data).encode("utf-8")
        self.assertRaises(webob.exc.HTTPBadRequest,
                          self.controller.create, req, self.share_id, data)
